#!/usr/bin/env python
# packet_out.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (packet_out.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
"""
.. module:: packet_out.

.. role:: red

BitDust packet_out() Automat

.. raw:: html

    <a href="packet_out.png" target="_blank">
    <img src="packet_out.png" style="max-width:100%;">
    </a>

EVENTS:
    * :red:`cancel`
    * :red:`failed`
    * :red:`inbox-packet`
    * :red:`item-cancelled`
    * :red:`items-sent`
    * :red:`nothing-to-send`
    * :red:`register-item`
    * :red:`remote-identity-on-hand`
    * :red:`request-failed`
    * :red:`response-timeout`
    * :red:`run`
    * :red:`unregister-item`
    * :red:`write-error`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from six.moves import map
from six.moves import range

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 16

_PacketLogFileEnabled = False

#------------------------------------------------------------------------------

import os
import time

#------------------------------------------------------------------------------

from twisted.internet import reactor  # @UnresolvedImport
from twisted.internet.defer import Deferred, CancelledError

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.automats import automat

from bitdust.p2p import commands
from bitdust.p2p import p2p_stats

from bitdust.lib import nameurl
from bitdust.lib import strng
from bitdust.lib import net_misc

from bitdust.system import tmpfile

from bitdust.contacts import contactsdb
from bitdust.contacts import identitycache

from bitdust.main import settings
from bitdust.main import config

from bitdust.transport import callback

from bitdust.userid import global_id
from bitdust.userid import id_url
from bitdust.userid import my_id

#------------------------------------------------------------------------------

_OutboxQueue = []
_PacketsCounter = 0

#------------------------------------------------------------------------------


def init():
    global _PacketLogFileEnabled
    _PacketLogFileEnabled = config.conf().getBool('logs/packet-enabled')


def shutdown():
    global _PacketLogFileEnabled
    _PacketLogFileEnabled = False


#------------------------------------------------------------------------------


def get_packets_counter():
    global _PacketsCounter
    return _PacketsCounter


def increment_packets_counter():
    global _PacketsCounter
    _PacketsCounter += 1


#------------------------------------------------------------------------------


def queue():
    global _OutboxQueue
    return _OutboxQueue


def create(outpacket, wide, callbacks, target=None, route=None, response_timeout=None, keep_alive=True, skip_ack=False):
    if _Debug:
        lg.out(
            _DebugLevel, 'packet_out.create [%s/%s/%s]:%s(%s) target=%r route=%r callbacks=%s' % (
                nameurl.GetName(outpacket.OwnerID),
                nameurl.GetName(outpacket.CreatorID),
                nameurl.GetName(outpacket.RemoteID),
                outpacket.Command,
                outpacket.PacketID,
                target,
                route,
                list(callbacks.keys()),
            )
        )
    p = PacketOut(outpacket, wide, callbacks, target, route, response_timeout, keep_alive, skip_ack=skip_ack)
    queue().append(p)
    p.automat('run')
    return p


#------------------------------------------------------------------------------


def search(proto, host, filename, remote_idurl=None):
    for p in queue():
        if p.filename != filename:
            continue
        for i in p.items:
            if i.proto == proto:
                if not remote_idurl:
                    return p, i
                if p.remote_idurl and id_url.is_cached(p.remote_idurl) and id_url.is_cached(remote_idurl):
                    if id_url.field(remote_idurl).to_bin() != id_url.field(p.remote_idurl).to_bin():
                        if _Debug:
                            lg.out(_DebugLevel, 'packet_out.search found a packet addressed to another user: %s != %s' % (p.remote_idurl, remote_idurl))
                            lg.args(_DebugLevel, proto=proto, host=host, filename=filename, route=p.route, outpacket=p.outpacket)
                        continue
                return p, i
    if _Debug:
        for p in queue():
            if p.filename:
                lg.out(_DebugLevel, '%s [%s]' % (os.path.basename(p.filename), ('|'.join(['%s:%s' % (i.proto, i.host) for i in p.items]))))
            else:
                lg.warn('%s was not initialized yet' % str(p))
    return None, None


def search_by_packet_id(packet_id):
    result = []
    for p in queue():
        if p.outpacket.PacketID.count(packet_id):
            result.append(p)
    if _Debug:
        lg.out(_DebugLevel, 'packet_out.search_by_packet_id %s:' % packet_id)
        lg.out(_DebugLevel, '%s' % ('        \n'.join(map(str, result))))
    return result


def search_many(
    proto=None,
    host=None,
    filename=None,
    command=None,
    remote_idurl=None,
    packet_id=None,
):
    results = []
    for p in queue():
        if remote_idurl and id_url.field(p.remote_idurl).to_bin() != id_url.field(remote_idurl).to_bin():
            continue
        if filename and p.filename != filename:
            continue
        if command and p.outpacket.Command != command:
            continue
        if packet_id and p.outpacket.PacketID != packet_id:
            continue
        for i in p.items:
            if proto and i.proto != proto:
                continue
            if host and i.host != host:
                continue
            results.append((p, i))
    if _Debug:
        lg.out(_DebugLevel, 'packet_out.search_many query: (%s, %s, %s, %s) found %d items : ' % (proto, host, filename, remote_idurl, len(results)))
        lg.out(_DebugLevel, '%s' % ('        \n'.join(map(str, results))))
    return results


def search_by_transfer_id(transfer_id):
    for p in queue():
        for i in p.items:
            if i.transfer_id and i.transfer_id == transfer_id:
                return p, i
    return None, None


def search_by_response_packet(newpacket=None, proto=None, host=None, outgoing_command=None, incoming_command=None, incoming_packet_id=None, incoming_owner_idurl=None, incoming_creator_idurl=None, incoming_remote_idurl=None):
    result = []
    if incoming_owner_idurl is None and newpacket:
        incoming_owner_idurl = newpacket.OwnerID
    if incoming_creator_idurl is None and newpacket:
        incoming_creator_idurl = newpacket.CreatorID
    if incoming_remote_idurl is None and newpacket:
        incoming_remote_idurl = newpacket.RemoteID
    if incoming_packet_id is None and newpacket:
        incoming_packet_id = newpacket.PacketID
    if incoming_command is None and newpacket:
        incoming_command = newpacket.Command
    if _Debug:
        lg.out(
            _DebugLevel, 'packet_out.search_by_response_packet for incoming [%s/%s/%s]:%s|%s@%s from [%s://%s]' % (
                nameurl.GetName(incoming_owner_idurl),
                nameurl.GetName(incoming_creator_idurl),
                nameurl.GetName(incoming_remote_idurl),
                outgoing_command,
                incoming_command,
                incoming_packet_id,
                proto,
                host,
            )
        )
    matching_packet_ids = []
    matching_packet_ids.append(incoming_packet_id.lower())
    if incoming_command and incoming_command in [
        commands.Data(),
        commands.Retrieve(),
    ] and id_url.is_cached(incoming_owner_idurl) and incoming_owner_idurl == my_id.getIDURL():
        my_rotated_idurls = id_url.list_known_idurls(my_id.getIDURL(), num_revisions=10, include_revisions=False)
        # TODO: my_rotated_idurls can be cached for optimization
        for another_idurl in my_rotated_idurls:
            another_packet_id = global_id.SubstitutePacketID(incoming_packet_id, idurl=another_idurl).lower()
            if another_packet_id not in matching_packet_ids:
                matching_packet_ids.append(another_packet_id)
    if len(matching_packet_ids) > 1:
        if _Debug:
            lg.dbg(_DebugLevel, 'multiple packet IDs expecting to match for that packet: %r' % matching_packet_ids)
    matching_packet_ids_count = 0
    matching_command_ack_count = 0
    for p in queue():
        if p.outpacket.PacketID.lower() not in matching_packet_ids:
            # PacketID of incoming packet not matching with that outgoing packet
            continue
        matching_packet_ids_count += 1
        if p.outpacket.PacketID != incoming_packet_id:
            lg.warn('packet ID in queue "almost" matching with incoming: %s ~ %s' % (p.outpacket.PacketID, incoming_packet_id))
        if outgoing_command is None and not commands.IsCommandAck(p.outpacket.Command, incoming_command):
            # this command must not be in the reply
            continue
        matching_command_ack_count += 1
        if outgoing_command is not None and outgoing_command != p.outpacket.Command:
            # just in case if we are looking for some specific outgoing command
            continue
        expected_recipient = [
            p.outpacket.RemoteID,
        ]
        if id_url.is_cached(p.outpacket.RemoteID) and id_url.is_cached(p.remote_idurl):
            if p.outpacket.RemoteID != id_url.field(p.remote_idurl):
                # for Retrieve() packets I expect response exactly from target node
                if p.outpacket.Command != commands.Retrieve():
                    # outgoing packet was addressed to another node, so that means we need to expect response from another node also
                    expected_recipient.append(id_url.field(p.remote_idurl))
        matched = False
        if not id_url.is_cached(incoming_remote_idurl):
            identitycache.start_one(incoming_remote_idurl)
        else:
            if id_url.is_in(incoming_owner_idurl, expected_recipient, as_field=False):
                if id_url.is_the_same(my_id.getIDURL(), incoming_remote_idurl):
                    if _Debug:
                        lg.out(_DebugLevel, 'packet_out.search_by_response_packet    matched with incoming owner: %s' % expected_recipient)
                    matched = True
            if not matched:
                if id_url.is_in(incoming_creator_idurl, expected_recipient, as_field=False):
                    if id_url.is_the_same(my_id.getIDURL(), incoming_remote_idurl):
                        if _Debug:
                            lg.out(_DebugLevel, 'packet_out.search_by_response_packet    matched with incoming creator: %s' % expected_recipient)
                        matched = True
            if not matched:
                if id_url.is_in(incoming_remote_idurl, expected_recipient, as_field=False):
                    if id_url.is_the_same(my_id.getIDURL(), incoming_owner_idurl) and incoming_command == commands.Data():
                        if _Debug:
                            lg.out(_DebugLevel, 'packet_out.search_by_response_packet    matched my own incoming Data with incoming remote: %s' % expected_recipient)
                        matched = True
        if matched:
            result.append(p)
            if _Debug:
                lg.out(
                    _DebugLevel, 'packet_out.search_by_response_packet        FOUND pending outbox [%s/%s/%s]:%s(%s) cb:%s' % (
                        nameurl.GetName(p.outpacket.OwnerID),
                        nameurl.GetName(p.outpacket.CreatorID),
                        nameurl.GetName(p.outpacket.RemoteID),
                        p.outpacket.Command,
                        p.outpacket.PacketID,
                        list(p.callbacks.keys()),
                    )
                )
    if len(result) == 0:
        if _Debug:
            lg.out(_DebugLevel, 'packet_out.search_by_response_packet        DID NOT FOUND pending packets in outbox queue matching incoming %r' % newpacket)
            lg.args(_DebugLevel, pkt_ids_count=matching_packet_ids_count, cmd_ack_count=matching_command_ack_count, matching_packet_ids=matching_packet_ids)
    return result


def search_similar_packets(outpacket):
    return search_many(
        command=outpacket.Command,
        packet_id=outpacket.PacketID,
        remote_idurl=outpacket.RemoteID,
    )


#------------------------------------------------------------------------------


def on_outgoing_packet_failed(result, *a, **kw):
    if _Debug:
        lg.args(_DebugLevel, result=result, args=a, kwargs=kw)
    if result.type == CancelledError:
        return None
    return result


#------------------------------------------------------------------------------


class WorkItem(object):

    def __init__(self, proto, host, size=0):
        self.proto = proto
        self.host = net_misc.pack_address(host)
        self.time = time.time()
        self.transfer_id = None
        self.status = None
        self.error_message = None
        self.bytes_sent = 0
        self.size = size

    def __repr__(self):
        return 'WorkItem(%s://%s|%d)' % (self.proto, self.host, self.size)


#------------------------------------------------------------------------------


class PacketOut(automat.Automat):

    """
    This class implements all the functionality of the ``packet_out()`` state
    machine.
    """

    MESSAGES = {
        'MSG_1': 'file in queue was cancelled',
        'MSG_2': 'sending file was cancelled',
        'MSG_3': 'response waiting were cancelled',
        'MSG_4': 'outgoing packet was cancelled',
        'MSG_5': 'pushing outgoing packet was cancelled',
    }

    def __init__(self, outpacket, wide, callbacks={}, target=None, route=None, response_timeout=None, keep_alive=True, skip_ack=False):
        self.outpacket = outpacket
        if self.outpacket.PacketID.count('&'):
            packet_label = self.outpacket.PacketID.replace(':', '').replace('/', '').replace('_', '').replace('&', '')
        else:
            packet_label = global_id.ParseGlobalID(self.outpacket.PacketID)['path']
            if not packet_label:
                packet_label = self.outpacket.PacketID.replace(':', '').replace('/', '').replace('_', '')
        self.wide = wide
        self.callbacks = {}
        self.caching_deferred = None
        self.finished_deferred = Deferred()
        self.finished_deferred.addErrback(on_outgoing_packet_failed)
        self.final_result = None
        self.description = self.outpacket.Command + '[' + self.outpacket.PacketID + ']'
        self.remote_idurl = id_url.field(target) if target else None
        self.route = route
        self.response_timeout = response_timeout
        if self.route and 'remoteid' in self.route:
            self.description = self.route.get('description', self.description)
            self.remote_idurl = id_url.field(self.route['remoteid'])
        if not self.remote_idurl:
            self.remote_idurl = self.outpacket.RemoteID
        if not self.remote_idurl:
            raise ValueError('outgoing packet %r did not define remote idurl' % outpacket)
        self.remote_name = nameurl.GetName(self.outpacket.RemoteID)
        if id_url.to_bin(self.remote_idurl) != self.outpacket.RemoteID.to_bin():
            self.label = 'out_%d_%s_%s_via_%s' % (
                get_packets_counter(),
                packet_label,
                self.remote_name,
                nameurl.GetName(self.remote_idurl),
            )
        else:
            self.label = 'out_%d_%s_%s' % (get_packets_counter(), packet_label, self.remote_name)
        self.keep_alive = keep_alive
        self.skip_ack = skip_ack
        automat.Automat.__init__(
            self,
            name=self.label,
            state='AT_STARTUP',
            debug_level=_DebugLevel,
            log_events=_Debug,
            log_transitions=_Debug,
            publish_events=False,
        )
        increment_packets_counter()
        for command, cb in callbacks.items():
            self.set_callback(command, cb)

    def __repr__(self):
        """
        Will return something like: "out_123_alice[Data(9999999999)](SENDING)".
        """
        packet_label = '?'
        if self.outpacket:
            packet_label = '%s@%s' % (
                self.outpacket.Command,
                self.outpacket.PacketID,
            )  # .replace(':', '').replace('/', '').replace('_', '')
        return '%s[%s](%s)' % (self.id, packet_label, self.state)

    def init(self):
        """
        Method to initialize additional variables and flags at creation of the
        state machine.
        """
        self.error_message = None
        self.time = time.time()
        self.description = self.outpacket.Command + '(' + self.outpacket.PacketID + ')'
        self.payloadsize = len(self.outpacket.Payload)
        last_modified_time = identitycache.GetLastModifiedTime(self.remote_idurl)
        self.remote_identity = None
        if last_modified_time and time.time() - last_modified_time < 5*60:
            # use known that identity from cache if we sure that it is fresh enough
            self.remote_identity = contactsdb.get_contact_identity(self.remote_idurl)
            if self.remote_identity:
                if _Debug:
                    lg.out(_DebugLevel, 'packet_out.init  found fresh and cached identity: %s' % self.remote_idurl)
            else:
                if _Debug:
                    lg.out(_DebugLevel, 'packet_out.init  did not found cached identity: %s' % self.remote_idurl)
        else:
            if last_modified_time:
                if _Debug:
                    lg.out(_DebugLevel, 'packet_out.init  cached identity copy is out-dated: %s' % self.remote_idurl)
            else:
                if _Debug:
                    lg.out(_DebugLevel, 'packet_out.init  no cached identity copy exist or last caching time unknown: %s' % self.remote_idurl)
        self.packetdata = None
        self.filename = None
        self.filesize = None
        self.items = []
        self.results = []
        self.response_packet = None
        self.response_info = None
        self.timeout = None
        if self.response_timeout:
            self.timers['response-timeout'] = (self.response_timeout, [
                'RESPONSE?',
            ])

    def msg(self, msgid, *args, **kwargs):
        return self.MESSAGES.get(msgid, '')

    def is_timed_out(self):
        if self.state == 'RESPONSE?':
            return False
        if self.time is None or self.timeout is None:
            return False
        return time.time() - self.time > self.timeout

    def set_callback(self, command, cb):
        if command not in self.callbacks.keys():
            self.callbacks[command] = []
        self.callbacks[command].append(cb)

    def A(self, event, *args, **kwargs):
        #---SENDING---
        if self.state == 'SENDING':
            if event == 'register-item':
                self.doSetTransferID(*args, **kwargs)
            elif event == 'cancel':
                self.state = 'CANCEL'
                self.doCancelItems(*args, **kwargs)
                self.doErrMsg(event, self.msg('MSG_2', *args, **kwargs))
                self.doReportCancelItems(*args, **kwargs)
                self.doPopItems(*args, **kwargs)
                self.doReportCancelled(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'item-cancelled' and self.isMoreItems(*args, **kwargs):
                self.doPopItem(*args, **kwargs)
                self.doReportItem(*args, **kwargs)
            elif (event == 'unregister-item' and self.isFailed(*args, **kwargs)) or (event == 'item-cancelled' and not self.isMoreItems(*args, **kwargs)):
                self.state = 'FAILED'
                self.doPopItem(*args, **kwargs)
                self.doReportItem(*args, **kwargs)
                self.doReportFailed(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'unregister-item' and not self.isFailed(*args, **kwargs) and self.isAckNeeded(*args, **kwargs):
                self.state = 'RESPONSE?'
                self.doPopItem(*args, **kwargs)
                self.doReportItem(*args, **kwargs)
            elif event == 'unregister-item' and not self.isFailed(*args, **kwargs) and not self.isAckNeeded(*args, **kwargs):
                self.state = 'SENT'
                self.doPopItem(*args, **kwargs)
                self.doReportItem(*args, **kwargs)
                self.doReportDoneNoAck(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
        #---AT_STARTUP---
        elif self.state == 'AT_STARTUP':
            if event == 'run' and self.isRemoteIdentityKnown(*args, **kwargs):
                self.state = 'ITEMS?'
                self.doInit(*args, **kwargs)
                self.Cancelled = False
                self.doReportStarted(*args, **kwargs)
                self.doSerializeAndWrite(*args, **kwargs)
                self.doPushItems(*args, **kwargs)
            elif event == 'run' and not self.isRemoteIdentityKnown(*args, **kwargs):
                self.state = 'CACHING'
                self.doInit(*args, **kwargs)
                self.doCacheRemoteIdentity(*args, **kwargs)
        #---CACHING---
        elif self.state == 'CACHING':
            if event == 'remote-identity-on-hand':
                self.state = 'ITEMS?'
                self.Cancelled = False
                self.doReportStarted(*args, **kwargs)
                self.doSerializeAndWrite(*args, **kwargs)
                self.doPushItems(*args, **kwargs)
            elif event == 'failed':
                self.state = 'FAILED'
                self.doReportFailed(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'cancel':
                self.state = 'CANCEL'
                self.doErrMsg(event, self.msg('MSG_4', *args, **kwargs))
                self.doReportCancelled(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
        #---ITEMS?---
        elif self.state == 'ITEMS?':
            if event == 'nothing-to-send' or event == 'write-error':
                self.state = 'FAILED'
                self.doReportFailed(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'items-sent' and not self.Cancelled:
                self.state = 'IN_QUEUE'
            elif event == 'cancel':
                self.Cancelled = True
            elif event == 'items-sent' and self.Cancelled:
                self.state = 'CANCEL'
                self.doCancelItems(*args, **kwargs)
                self.doErrMsg(event, self.msg('MSG_5', *args, **kwargs))
                self.doReportCancelItems(*args, **kwargs)
                self.doPopItems(*args, **kwargs)
                self.doReportCancelled(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
        #---IN_QUEUE---
        elif self.state == 'IN_QUEUE':
            if event == 'item-cancelled' and self.isMoreItems(*args, **kwargs):
                self.doPopItem(*args, **kwargs)
            elif event == 'register-item':
                self.state = 'SENDING'
                self.doSetTransferID(*args, **kwargs)
            elif event == 'cancel':
                self.state = 'CANCEL'
                self.doCancelItems(*args, **kwargs)
                self.doErrMsg(event, self.msg('MSG_1', *args, **kwargs))
                self.doReportCancelItems(*args, **kwargs)
                self.doPopItems(*args, **kwargs)
                self.doReportCancelled(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'item-cancelled' and not self.isMoreItems(*args, **kwargs):
                self.state = 'FAILED'
                self.doPopItem(*args, **kwargs)
                self.doReportItem(*args, **kwargs)
                self.doReportFailed(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
        #---RESPONSE?---
        elif self.state == 'RESPONSE?':
            if event == 'cancel':
                self.state = 'CANCEL'
                self.doErrMsg(event, self.msg('MSG_3', *args, **kwargs))
                self.doReportCancelItems(*args, **kwargs)
                self.doReportCancelled(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'inbox-packet' and self.isResponse(*args, **kwargs):
                self.state = 'SENT'
                self.doSaveResponse(*args, **kwargs)
                self.doReportResponse(*args, **kwargs)
                self.doReportDoneWithAck(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'unregister-item' or event == 'item-cancelled':
                self.doPopItem(*args, **kwargs)
                self.doReportItem(*args, **kwargs)
            elif (event == 'response-timeout' or event == 'request-failed') and not self.isDataExpected(*args, **kwargs):
                self.state = 'SENT'
                self.doReportTimeOut(*args, **kwargs)
                self.doReportDoneNoAck(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif (event == 'response-timeout' or event == 'request-failed') and self.isDataExpected(*args, **kwargs):
                self.state = 'FAILED'
                self.doReportFailed(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
        #---FAILED---
        elif self.state == 'FAILED':
            pass
        #---SENT---
        elif self.state == 'SENT':
            pass
        #---CANCEL---
        elif self.state == 'CANCEL':
            pass
        return None

    def isRemoteIdentityKnown(self, *args, **kwargs):
        """
        Condition method.
        """
        return self.remote_identity is not None

    def isAckNeeded(self, *args, **kwargs):
        """
        Condition method.
        """
        if self.skip_ack:
            return False
        if commands.IsReplyExpected(self.outpacket.Command):
            if not self.response_timeout and self.outpacket.Command == commands.Message():
                # an exception for sending back a Message() packet with list of archived messages
                # TODO: more elegant solution to be found
                return False
            return True
        return len(self.callbacks.get(commands.Ack(), [])) + len(self.callbacks.get(commands.Fail(), [])) > 0

    def isFailed(self, *args, **kwargs):
        """
        Condition method.
        """
        return args[0][1] != 'finished'

    def isMoreItems(self, *args, **kwargs):
        """
        Condition method.
        """
        return len(self.items) > 1

    def isResponse(self, *args, **kwargs):
        """
        Condition method.
        """
        newpacket, _ = args[0]
        if len(self.callbacks.get(newpacket.Command, [])) > 0:
            return True
        if not commands.IsCommandAck(self.outpacket.Command, newpacket.Command):
            return False
        return True

    def isDataExpected(self, *args, **kwargs):
        """
        Condition method.
        """
        return len(self.callbacks.get(commands.Data(), [])) > 0

    def doInit(self, *args, **kwargs):
        """
        Action method.
        """
        if self in self.outpacket.Packets:
            lg.warn('packet_out already connected to the packet')
        else:
            self.outpacket.Packets.append(self)

    def doCacheRemoteIdentity(self, *args, **kwargs):
        """
        Action method.
        """
        self.caching_deferred = identitycache.immediatelyCaching(self.remote_idurl)
        self.caching_deferred.addCallback(self._on_remote_identity_cached)
        self.caching_deferred.addErrback(self._on_remote_identity_cache_failed)
        self.caching_deferred.addTimeout(60, clock=reactor)

    def doSerializeAndWrite(self, *args, **kwargs):
        """
        Action method.
        """
        # serialize and write packet on disk
        a_packet = self.outpacket
        if self.route:
            a_packet = self.route.get('packet', a_packet)
        try:
            fileno, self.filename = tmpfile.make('outbox', extension='.out')
            self.packetdata = a_packet.Serialize()
            os.write(fileno, self.packetdata)
            os.close(fileno)
            self.filesize = len(self.packetdata)
            if self.filesize < 1024*10:
                if self.response_timeout:
                    self.timeout = self.response_timeout
                else:
                    self.timeout = settings.P2PTimeOut()
            elif self.filesize > 1024*1024:
                self.timeout = 5 + int(self.filesize/float(settings.SendingSpeedLimit()))
            else:
                self.timeout = 300
        except:
            lg.exc()
            self.packetdata = None
            reactor.callLater(0, self.automat, 'write-error')  # @UndefinedVariable

    def doPushItems(self, *args, **kwargs):
        """
        Action method.
        """
        self._push()

    def doPopItem(self, *args, **kwargs):
        """
        Action method.
        """
        self._pop(args[0])

    def doPopItems(self, *args, **kwargs):
        """
        Action method.
        """
        self.items = []

    def doSetTransferID(self, *args, **kwargs):
        """
        Action method.
        """
        ok = False
        proto, host, _, transfer_id = args[0]
        for i in range(len(self.items)):
            if self.items[i].proto == proto:
                self.items[i].transfer_id = transfer_id
                if _Debug:
                    lg.out(_DebugLevel, 'packet_out.doSetTransferID  %r:%r = %r' % (proto, host, transfer_id))
                ok = True
        if not ok:
            lg.warn('not found item for %r:%r' % (proto, host))

    def doSaveResponse(self, *args, **kwargs):
        """
        Action method.
        """
        self.response_packet, self.response_info = args[0]

    def doCancelItems(self, *args, **kwargs):
        """
        Action method.
        """
        from bitdust.transport import gateway
        for i in self.items:
            t = gateway.transports().get(i.proto, None)
            if t:
                if i.transfer_id:
                    t.call('cancel_file_sending', i.transfer_id)
                t.call('cancel_outbox_file', i.host, self.filename)

    def doReportStarted(self, *args, **kwargs):
        """
        Action method.
        """
        handled = callback.run_outbox_callbacks(self)
        if not handled:
            pass

    def doReportItem(self, *args, **kwargs):
        """
        Action method.
        """
        if not self.popped_item:
            raise Exception('current outgoing item not exist')
        if _PacketLogFileEnabled:
            if self.popped_item.status == 'finished':
                lg.out(
                    0,
                    '\033[0;49;90mSENT %d bytes to %s://%s TID:%s\033[0m' % (
                        self.popped_item.bytes_sent,
                        strng.to_text(self.popped_item.proto),
                        strng.to_text(self.popped_item.host),
                        self.popped_item.transfer_id,
                    ),
                    log_name='packet',
                    showtime=True,
                )
            else:
                lg.out(
                    0,
                    '\033[0;49;91mFAILED %d bytes to %s://%s with status=%r TID:%s\033[0m' % (
                        self.popped_item.bytes_sent,
                        strng.to_text(self.popped_item.proto),
                        strng.to_text(self.popped_item.host),
                        self.popped_item.status,
                        self.popped_item.transfer_id,
                    ),
                    log_name='packet',
                    showtime=True,
                )
        p2p_stats.count_outbox(self.remote_idurl, self.popped_item.proto, self.popped_item.status, self.popped_item.bytes_sent)
        callback.run_finish_file_sending_callbacks(self, self.popped_item, self.popped_item.status, self.popped_item.bytes_sent, self.popped_item.error_message)
        if self.popped_item.status == 'failed':
            for cb in self.callbacks.pop('item-failed', []):
                cb(self, self.popped_item)
        else:
            for cb in self.callbacks.pop('item-sent', []):
                cb(self, self.popped_item)
        self.popped_item = None

    def doReportCancelItems(self, *args, **kwargs):
        """
        Action method.
        """
        for item in self.results:
            if _PacketLogFileEnabled:
                lg.out(
                    0,
                    '\033[0;49;90mOUT CANCELED %s://%s TID:%s\033[0m' % (strng.to_text(item.proto), strng.to_text(item.host), item.transfer_id),
                    log_name='packet',
                    showtime=True,
                )
            p2p_stats.count_outbox(self.remote_idurl, item.proto, 'failed', 0)
            callback.run_finish_file_sending_callbacks(self, item, 'failed', 0, self.error_message)

    def doReportResponse(self, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.out(_DebugLevel, 'packet_out.doReportResponse %d callbacks known' % len(self.callbacks))
        for cb in self.callbacks.pop(self.response_packet.Command, []):
            if _Debug:
                lg.out(_DebugLevel, '        calling to %r with %r' % (cb, self.response_packet))
            try:
                cb(self.response_packet, self.response_info)
            except:
                lg.exc()

    def doReportTimeOut(self, *args, **kwargs):
        """
        Action method.
        """
        self.final_result = 'timeout'
        if _PacketLogFileEnabled:
            lg.out(
                0,
                '\033[1;49;91mOUT TIMEOUT %s(%s) sending from %s to %s\033[0m' % (
                    self.outpacket.Command,
                    self.outpacket.PacketID,
                    global_id.UrlToGlobalID(self.outpacket.CreatorID),
                    global_id.UrlToGlobalID(self.remote_idurl),
                ),
                log_name='packet',
                showtime=True,
            )
        for cb in self.callbacks.pop(None, []):
            cb(self)
        for cb in self.callbacks.pop('timeout', []):
            cb(self, 'timeout')

    def doReportDoneWithAck(self, *args, **kwargs):
        """
        Action method.
        """
        self.final_result = 'finished'
        if _PacketLogFileEnabled:
            newpacket, _ = args[0]
            if newpacket.Command in [
                commands.Fail(),
            ]:
                lg.out(
                    0,
                    '                \033[0;49;31mRECEIVE %s on %s(%s) with %s bytes from %s to %s TID:%r\033[0m' % (
                        newpacket.Command,
                        self.outpacket.Command,
                        self.outpacket.PacketID,
                        self.filesize or '?',
                        global_id.UrlToGlobalID(self.outpacket.CreatorID),
                        global_id.UrlToGlobalID(self.remote_idurl),
                        [i.transfer_id for i in self.results],
                    ),
                    log_name='packet',
                    showtime=True,
                )
            else:
                lg.out(
                    0,
                    '                \033[1;49;96mRECEIVE %s on %s(%s) with %s bytes from %s to %s TID:%r\033[0m' % (
                        newpacket.Command,
                        self.outpacket.Command,
                        self.outpacket.PacketID,
                        self.filesize or '?',
                        global_id.UrlToGlobalID(self.outpacket.CreatorID),
                        global_id.UrlToGlobalID(self.remote_idurl),
                        [i.transfer_id for i in self.results],
                    ),
                    log_name='packet',
                    showtime=True,
                )
        callback.run_queue_item_status_callbacks(self, 'finished', '')
        for cb in self.callbacks.pop('acked', []):
            cb(self, 'finished')
        if not self.finished_deferred.called:
            self.finished_deferred.callback(self)

    def doReportDoneNoAck(self, *args, **kwargs):
        """
        Action method.
        """
        self.final_result = 'finished_no_ack'
        if _PacketLogFileEnabled:
            lg.out(
                0,
                '\033[0;49;95mOUT %s(%s) with %s bytes from %s to %s TID:%r\033[0m' % (
                    self.outpacket.Command,
                    self.outpacket.PacketID,
                    self.filesize or '?',
                    global_id.UrlToGlobalID(self.outpacket.CreatorID),
                    global_id.UrlToGlobalID(self.remote_idurl),
                    [i.transfer_id for i in self.results],
                ),
                log_name='packet',
                showtime=True,
            )
        if (args and args[0]) or self.skip_ack:
            callback.run_queue_item_status_callbacks(self, 'finished', '')
        else:
            callback.run_queue_item_status_callbacks(self, 'finished', 'unanswered')
        for cb in self.callbacks.pop('sent', []):
            cb(self, 'finished')
        if not self.finished_deferred.called:
            self.finished_deferred.callback(self)

    def doReportFailed(self, *args, **kwargs):
        """
        Action method.
        """
        try:
            msg = str(args[0][-1])
        except:
            msg = 'failed'
        self.final_result = 'failed'
        if _PacketLogFileEnabled:
            lg.out(
                0,
                '\033[0;49;91mOUT FAILED %s(%s) with %s bytes from %s to %s TID:%r : %s\033[0m' % (
                    self.outpacket.Command,
                    self.outpacket.PacketID,
                    self.filesize or '?',
                    global_id.UrlToGlobalID(self.outpacket.CreatorID),
                    global_id.UrlToGlobalID(self.remote_idurl),
                    [i.transfer_id for i in self.results],
                    msg,
                ),
                log_name='packet',
                showtime=True,
            )
        callback.run_queue_item_status_callbacks(self, 'failed', msg)
        for cb in self.callbacks.pop('failed', []):
            cb(self, msg)
        if not self.finished_deferred.called:
            self.finished_deferred.callback(self)

    def doReportCancelled(self, *args, **kwargs):
        """
        Action method.
        """
        msg = 'cancelled'
        if args and args[0]:
            msg = str(args[0])
        else:
            msg = 'cancelled'
        self.final_result = msg
        if self.final_result == 'timeout':
            if _PacketLogFileEnabled:
                lg.out(
                    0,
                    '\033[0;49;97mOUT CANCELED %s(%s) after TIMEOUT with %s bytes from %s to %s TID:%r\033[0m' % (
                        self.outpacket.Command,
                        self.outpacket.PacketID,
                        self.filesize or '?',
                        global_id.UrlToGlobalID(self.outpacket.CreatorID),
                        global_id.UrlToGlobalID(self.remote_idurl),
                        [i.transfer_id for i in self.results],
                    ),
                    log_name='packet',
                    showtime=True,
                )
            for cb in self.callbacks.pop(None, []):
                cb(self)
            for cb in self.callbacks.pop('timeout', []):
                cb(self, 'timeout')
        else:
            if _PacketLogFileEnabled:
                lg.out(
                    0,
                    '\033[0;49;97mOUT CANCELED %s(%s) with %s bytes from %s to %s TID:%r : %s\033[0m' % (
                        self.outpacket.Command,
                        self.outpacket.PacketID,
                        self.filesize or '?',
                        global_id.UrlToGlobalID(self.outpacket.CreatorID),
                        global_id.UrlToGlobalID(self.remote_idurl),
                        [i.transfer_id for i in self.results],
                        msg,
                    ),
                    log_name='packet',
                    showtime=True,
                )
            callback.run_queue_item_status_callbacks(self, 'cancelled', msg)
            for cb in self.callbacks.pop('cancelled', []):
                cb(self, msg)
            for cb in self.callbacks.pop(None, []):
                cb(self)
            if not self.finished_deferred.called:
                self.finished_deferred.callback(self)

    def doErrMsg(self, event, *args, **kwargs):
        """
        Action method.
        """
        if event.count('timer'):
            self.error_message = 'timeout responding from remote side'
        else:
            self.error_message = args[0]

    def doDestroyMe(self, *args, **kwargs):
        """
        Remove all references to the state machine object to destroy it.
        """
        queue().remove(self)
        if self not in self.outpacket.Packets:
            lg.warn('packet_out not connected to the packet')
        else:
            self.outpacket.Packets.remove(self)
        self.outpacket = None
        self.remote_identity = None
        if self.caching_deferred and not self.caching_deferred.called:
            self.caching_deferred.cancel()
        self.caching_deferred = None
        self.callbacks.clear()
        if self.finished_deferred and not self.finished_deferred.called:
            self.finished_deferred.cancel()
        self.destroy()

    def _on_remote_identity_cached(self, xmlsrc):
        self.remote_identity = contactsdb.get_contact_identity(self.remote_idurl)
        if self.remote_identity is None:
            reactor.callLater(0, self.automat, 'failed')  # @UndefinedVariable
        else:
            reactor.callLater(0, self.automat, 'remote-identity-on-hand')  # @UndefinedVariable
        return xmlsrc

    def _on_remote_identity_cache_failed(self, err):
        lg.warn('%s : %s' % (repr(self), str(err)))
        if self.outpacket:
            reactor.callLater(0, self.automat, 'failed')  # @UndefinedVariable
        return None

    def _push(self):
        from bitdust.transport import gateway
        if self.route and 'proto' in self.route and 'host' in self.route and 'remoteid' in self.route:
            # if this packet is routed - send directly to the host specified in the route info
            proto = strng.to_text(self.route['proto'])
            host = strng.to_bin(self.route['host'])
            if not gateway.send_file(
                strng.to_bin(self.route['remoteid']),
                proto,
                host,
                self.filename,
                self.description,
                self,
            ):
                self.automat('nothing-to-send')
                if _PacketLogFileEnabled:
                    lg.out(0, '\033[0;49;97mSKIP sending routed %r : filtered out\033[0m' % self, log_name='packet', showtime=True)
                return
            self.items.append(WorkItem(proto, host, self.filesize))
            if _PacketLogFileEnabled:
                lg.out(
                    0,
                    '\033[0;49;90mPUSH %d bytes to %s://%s ROUTED\033[0m' % (self.filesize, strng.to_text(proto), strng.to_text(host)),
                    log_name='packet',
                    showtime=True,
                )
            self.automat('items-sent')
            return
        # get info about his local IP
        localIP = identitycache.GetLocalIP(self.remote_idurl)
        workitem_sent = False
        if self.wide:
            # send to all his contacts
            for contactmethod in self.remote_identity.getContacts():
                proto, host = nameurl.IdContactSplit(contactmethod)
                proto = strng.to_text(proto)
                host = strng.to_bin(host)
                if host.strip() and \
                        settings.transportIsEnabled(proto) and \
                        settings.transportSendingIsEnabled(proto) and \
                        gateway.is_installed(proto) and \
                        gateway.can_send(proto):
                    if proto == 'tcp' and localIP:
                        host = localIP
                    if not gateway.send_file(
                        self.remote_idurl.to_bin(),
                        proto,
                        host,
                        self.filename,
                        self.description,
                        self,
                    ):
                        continue
                    self.items.append(WorkItem(proto, host, self.filesize))
                    workitem_sent = True
                    if _PacketLogFileEnabled:
                        lg.out(
                            0,
                            '\033[0;49;90mPUSH %d bytes to %s://%s\033[0m' % (self.filesize, strng.to_text(proto), strng.to_text(host)),
                            log_name='packet',
                            showtime=True,
                        )
            if not workitem_sent:
                self.automat('nothing-to-send')
                lg.warn('(wide) no supported protocols with %s' % self.remote_idurl)
                if _PacketLogFileEnabled:
                    lg.out(0, '\033[0;49;97mSKIP wide sending %r : no supported protocols\033[0m' % self, log_name='packet', showtime=True)
            else:
                self.automat('items-sent')
            return
        # send to one of his contacts,
        # now need to decide which transport to use
        # let's prepare his contacts first
        byproto = self.remote_identity.getContactsByProto()
        tcp_contact = None
        if settings.enableTCP() and settings.enableTCPsending():
            tcp_contact = byproto.get('tcp', None)
        udp_contact = None
        if settings.enableUDP() and settings.enableUDPsending():
            udp_contact = byproto.get('udp', None)
        http_contact = None
        if settings.enableHTTP() and settings.enableHTTPsending():
            http_contact = byproto.get('http', None)
        proxy_contact = None
        if settings.enablePROXY() and settings.enablePROXYsending():
            proxy_contact = byproto.get('proxy', None)
        working_protos = p2p_stats.peers_protos().get(self.remote_idurl, set())
        # tcp seems to be the most stable proto
        # now let's check if we know his local IP and
        # he enabled tcp in his settings to be able to receive packets from others
        # try to send to his local IP first, not external
        if tcp_contact and localIP:
            if gateway.is_installed('tcp') and gateway.can_send(proto):
                proto, host, port, _ = nameurl.UrlParse(tcp_contact)
                if port:
                    host = localIP + ':' + str(port)
                proto = strng.to_text(proto)
                host = strng.to_bin(host)
                if gateway.send_file(self.remote_idurl.to_bin(), proto, host, self.filename, self.description, self):
                    self.items.append(WorkItem(proto, host, self.filesize))
                    if _PacketLogFileEnabled:
                        lg.out(
                            0,
                            '\033[0;49;90mPUSH %d bytes to %s://%s\033[0m' % (self.filesize, strng.to_text(proto), strng.to_text(host)),
                            log_name='packet',
                            showtime=True,
                        )
                    self.automat('items-sent')
                    return
        # tcp is the best proto - if it is working - this is the best case!!!
        if tcp_contact and 'tcp' in working_protos:
            proto, host, port, _ = nameurl.UrlParse(tcp_contact)
            if host.strip() and gateway.is_installed(proto) and gateway.can_send(proto):
                if port:
                    host = host + ':' + str(port)
                proto = strng.to_text(proto)
                host = strng.to_bin(host)
                if gateway.send_file(self.remote_idurl.to_bin(), proto, host, self.filename, self.description):
                    self.items.append(WorkItem(proto, host, self.filesize))
                    if _PacketLogFileEnabled:
                        lg.out(
                            0,
                            '\033[0;49;90mPUSH %d bytes to %s://%s\033[0m' % (self.filesize, strng.to_text(proto), strng.to_text(host)),
                            log_name='packet',
                            showtime=True,
                        )
                    self.automat('items-sent')
                    return
        # udp contact
        if udp_contact and 'udp' in working_protos:
            proto, host = nameurl.IdContactSplit(udp_contact)
            if host.strip() and gateway.is_installed('udp') and gateway.can_send(proto):
                proto = strng.to_text(proto)
                host = strng.to_bin(host)
                if gateway.send_file(self.remote_idurl.to_bin(), proto, host, self.filename, self.description, self):
                    self.items.append(WorkItem(proto, host, self.filesize))
                    if _PacketLogFileEnabled:
                        lg.out(
                            0,
                            '\033[0;49;90mPUSH %d bytes to %s://%s\033[0m' % (self.filesize, strng.to_text(proto), strng.to_text(host)),
                            log_name='packet',
                            showtime=True,
                        )
                    self.automat('items-sent')
                    return
        # http contact
        if http_contact and 'http' in working_protos:
            proto, host, port, _ = nameurl.UrlParse(http_contact)
            if host.strip() and gateway.is_installed(proto) and gateway.can_send(proto):
                if port:
                    host = host + ':' + str(port)
                proto = strng.to_text(proto)
                host = strng.to_bin(host)
                if gateway.send_file(self.remote_idurl.to_bin(), proto, host, self.filename, self.description, self):
                    self.items.append(WorkItem(proto, host, self.filesize))
                    if _PacketLogFileEnabled:
                        lg.out(
                            0,
                            '\033[0;49;90mPUSH %d bytes to %s://%s\033[0m' % (self.filesize, strng.to_text(proto), strng.to_text(host)),
                            log_name='packet',
                            showtime=True,
                        )
                    self.automat('items-sent')
                    return
        # proxy contact - he may use other node to receive and send packets
        if proxy_contact and 'proxy' in working_protos:
            proto, host = nameurl.IdContactSplit(proxy_contact)
            if host.strip() and gateway.is_installed('proxy') and gateway.can_send(proto):
                proto = strng.to_text(proto)
                host = strng.to_bin(host)
                if gateway.send_file(self.remote_idurl.to_bin(), proto, host, self.filename, self.description, self):
                    self.items.append(WorkItem(proto, host, self.filesize))
                    if _PacketLogFileEnabled:
                        lg.out(
                            0,
                            '\033[0;49;90mPUSH %d bytes to %s://%s\033[0m' % (self.filesize, strng.to_text(proto), strng.to_text(host)),
                            log_name='packet',
                            showtime=True,
                        )
                    self.automat('items-sent')
                    return
        # finally use the first proto we supported if we can not find the best preferable method
        for contactmethod in self.remote_identity.getContacts():
            proto, host, port, _ = nameurl.UrlParse(contactmethod)
            if port:
                host = host + ':' + str(port)
            # if method exist but empty - don't use it
            if host.strip():
                # try sending with tcp even if it is switched off in the settings
                if gateway.is_installed(proto) and gateway.can_send(proto):
                    if settings.enableTransport(proto) and settings.transportSendingIsEnabled(proto):
                        if gateway.send_file(self.remote_idurl.to_bin(), strng.to_text(proto), strng.to_bin(host), self.filename, self.description, self):
                            self.items.append(WorkItem(strng.to_text(proto), strng.to_bin(host), self.filesize))
                            if _PacketLogFileEnabled:
                                lg.out(
                                    0,
                                    '\033[0;49;90mPUSH %d bytes to %s://%s\033[0m' % (self.filesize, strng.to_text(proto), strng.to_text(host)),
                                    log_name='packet',
                                    showtime=True,
                                )
                            self.automat('items-sent')
                            return
        self.automat('nothing-to-send')
        lg.warn('no supported protocols with %s : %s %s %s, byproto:%s' % (self.remote_idurl, tcp_contact, udp_contact, working_protos, str(byproto)))
        if _PacketLogFileEnabled:
            lg.out(0, '\033[0;49;97mSKIP sending %r : no supported protocols\033[0m' % self, log_name='packet', showtime=True)

    def _pop(self, packet_args):
        self.popped_item = None
        if len(packet_args) == 4:
            transfer_id, status, size, error_message = packet_args
            for i in self.items:
                if i.transfer_id and i.transfer_id == transfer_id:
                    self.items.remove(i)
                    i.status = status
                    i.error_message = error_message
                    i.bytes_sent = size
                    self.results.append(i)
                    self.popped_item = i
                    if _PacketLogFileEnabled:
                        lg.out(
                            0,
                            '\033[0;49;90mPOP %d bytes to %s://%s TID=%r status=%r\033[0m' % (size, strng.to_text(i.proto), strng.to_text(i.host), i.transfer_id, status),
                            log_name='packet',
                            showtime=True,
                        )
                    break
        elif len(packet_args) == 6:
            proto, host, _, size, _, err_msg = packet_args
            for i in self.items:
                if i.proto == proto and i.host == host:
                    self.items.remove(i)
                    i.status = 'failed'
                    i.error_message = err_msg
                    i.bytes_sent = size
                    self.results.append(i)
                    self.popped_item = i
                    if _PacketLogFileEnabled:
                        lg.out(
                            0,
                            '\033[0;49;90mPOP %d bytes to %s://%s TID=%r status=%r\033[0m' % (size, strng.to_text(i.proto), strng.to_text(i.host), i.transfer_id, i.status),
                            log_name='packet',
                            showtime=True,
                        )
                    break
        if not self.popped_item:
            raise Exception('failed to populate active item')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2011-2013 Codernity (http://codernity.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import absolute_import
import os
import marshal

import struct
import shutil

from bitdust_forks.CodernityDB.storage import IU_Storage, DummyStorage
import six

try:
    from CodernityDB import __version__
except ImportError:
    from .__init__ import __version__


import io


class IndexException(Exception):
    pass


class IndexNotFoundException(IndexException):
    pass


class ReindexException(IndexException):
    pass


class TryReindexException(ReindexException):
    pass


class ElemNotFound(IndexException):
    pass


class DocIdNotFound(ElemNotFound):
    pass


class IndexConflict(IndexException):
    pass


class IndexPreconditionsException(IndexException):
    pass


class Index(object):

    __version__ = __version__

    custom_header = ""  # : use it for imports required by your index

    def __init__(self,
                 db_path,
                 name):
        self.name = name
        self._start_ind = 500
        self.db_path = db_path

    def open_index(self):
        if not os.path.isfile(os.path.join(self.db_path, self.name + '_buck')):
            raise IndexException("Doesn't exists")
        self.buckets = io.open(
            os.path.join(self.db_path, self.name + "_buck"), 'r+b', buffering=0)
        self._fix_params()
        self._open_storage()

    def _close(self):
        self.buckets.close()
        self.storage.close()

    def close_index(self):
        self.flush()
        self.fsync()
        self._close()

    def create_index(self):
        raise NotImplementedError()

    def _fix_params(self):
        self.buckets.seek(0)
        props = marshal.loads(self.buckets.read(self._start_ind))
        for k, v in six.iteritems(props):
            self.__dict__[k] = v
        self.buckets.seek(0, 2)

    def _save_params(self, in_params={}):
        self.buckets.seek(0)
        props = marshal.loads(self.buckets.read(self._start_ind))
        props.update(in_params)
        self.buckets.seek(0)
        data = marshal.dumps(props)
        if len(data) > self._start_ind:
            raise IndexException("To big props")
        self.buckets.write(data)
        self.flush()
        self.buckets.seek(0, 2)
        self.__dict__.update(props)

    def _open_storage(self, *args, **kwargs):
        pass

    def _create_storage(self, *args, **kwargs):
        pass

    def _destroy_storage(self, *args, **kwargs):
        self.storage.destroy()

    def _find_key(self, key):
        raise NotImplementedError()

    def update(self, doc_id, key, start, size):
        raise NotImplementedError()

    def insert(self, doc_id, key, start, size):
        raise NotImplementedError()

    def get(self, key):
        raise NotImplementedError()

    def get_many(self, key, start_from=None, limit=0):
        raise NotImplementedError()

    def all(self, start_pos):
        raise NotImplementedError()

    def delete(self, key, start, size):
        raise NotImplementedError()

    def make_key_value(self, data):
        raise NotImplementedError()

    def make_key(self, data):
        raise NotImplementedError()

    def compact(self, *args, **kwargs):
        raise NotImplementedError()

    def destroy(self, *args, **kwargs):
        self._close()
        bucket_file = os.path.join(self.db_path, self.name + '_buck')
        os.unlink(bucket_file)
        self._destroy_storage()
        self._find_key.clear()

    def flush(self):
        try:
            self.buckets.flush()
            self.storage.flush()
        except:
            pass

    def fsync(self):
        try:
            os.fsync(self.buckets.fileno())
            self.storage.fsync()
        except:
            pass

    def update_with_storage(self, doc_id, key, value):
        if value:
            start, size = self.storage.insert(value)
        else:
            start = 1
            size = 0
        return self.update(doc_id, key, start, size)

    def insert_with_storage(self, doc_id, key, value):
        if value:
            start, size = self.storage.insert(value)
        else:
            start = 1
            size = 0
        return self.insert(doc_id, key, start, size)



#!/usr/bin/python
#http_node.py
#
#
#    Copyright BitDust, 2006
#    Use of this software constitutes acceptance of the Terms of Use
#      http://datahaven.net/terms_of_use.html
#    All rights reserved.
#
#

from __future__ import absolute_import
from __future__ import print_function
import os
import sys
import base64
import time

if __name__ == '__main__':
    dirpath = os.path.dirname(os.path.abspath(sys.argv[0]))
    sys.path.insert(0, os.path.abspath('datahaven'))
    sys.path.insert(0, os.path.abspath(os.path.join(dirpath, '..')))
    sys.path.insert(0, os.path.abspath(os.path.join(dirpath, '..', '..')))

try:
    from twisted.internet import reactor  # @UnresolvedImport
except:
    sys.exit('Error initializing twisted.internet.reactor in http_node.py')

from twisted.internet.defer import Deferred, succeed
from twisted.web import server, resource
from twisted.web.client import HTTPClientFactory

import misc
import dhnio
import dhnnet
import transport_control
import contacts
import nameurl
import settings
import identitycache
import tmpfile

from bitdust.logs import lg

_Outbox = {}
_Contacts = {}
_ReceivingLoop = None
_ServerListener = None
_LastPingTimeDict = {}
_PingDelayDict = {}
_ConnectionsDict = {}
_CurrentDelay = 5

#-------------------------------------------------------------------------------


def send(idurl, filename):
    lg.out(12, 'http_node.send to %s %s' % (idurl, filename))
    global _Outbox
    if idurl not in _Outbox:
        _Outbox[idurl] = []
    _Outbox[idurl].append(filename)
    #we want to keep only 10 last files.
    if len(_Outbox[idurl]) > 10:
        lostedfilename = _Outbox[idurl].pop(0)
        transport_control.sendStatusReport(
            'unknown',
            lostedfilename,
            'failed',
            'http',
        )


class SenderServer(resource.Resource):
    isLeaf = True

    def render_POST(self, request):
        global _Outbox
        idurl = request.getHeader('idurl')
        if idurl is None:
            return ''
        lg.out(14, 'http_node.SenderServer.render connection from ' + idurl)
        if not idurl in list(_Outbox.keys()):
            return ''
        r = ''
        for filename in _Outbox[idurl]:
            if not os.path.isfile(filename):
                continue
            if not os.access(filename, os.R_OK):
                continue
            src = dhnio.ReadBinaryFile(filename)
            if not src:
                continue
            src64 = base64.b64encode(src)
            r += src64 + '\n'
            lg.out(12, 'http_node.SenderServer.render sent %s to %s' % (filename, idurl))
            #TODO request.getPeer()
            transport_control.sendStatusReport(
                request.getClient(),
                filename,
                'finished',
                'http',
            )
        _Outbox.pop(idurl, None)
        return r


def start_http_server(port):
    global _ServerListener
    lg.out(6, 'http_node.start_http_server going to listen on port ' + str(port))
    if _ServerListener is not None:
        lg.out(8, 'http_node.start_http_server is already started')
        #        return _ServerListener
        return succeed(_ServerListener)

    def _try_listening(port, count):
        global _ServerListener
        lg.out(12, 'http_node.start_http_server count=%d' % count)
        site = server.Site(SenderServer())
        try:
            _ServerListener = reactor.listenTCP(int(port), site)
        except:
            _ServerListener = None
        return _ServerListener

    def _loop(port, result, count):
        l = _try_listening(port, count)
        if l is not None:
            lg.out(8, 'http_node.start_http_server started on port ' + str(port))
            result.callback(l)
            return
        if count > 10:
            lg.out(1, 'http_node.start_http_server WARNING port %s is busy!' % str(port))
            result.errback(None)
            return
        reactor.callLater(10, _loop, port, result, count + 1)

    res = Deferred()
    _loop(port, res, 0)
    return res


def stop_http_server():
    global _ServerListener
    lg.out(6, 'http_node.stop_http_server')
    if _ServerListener is None:
        lg.out(8, 'http_node.stop_http_server _ServerListener is None')
        d = Deferred()
        d.callback('')
        return d
    d = _ServerListener.stopListening()
    _ServerListener = None
    return d


class TransportHTTPClientFactory(HTTPClientFactory):
    pass


class TransportHTTPProxyClientFactory(HTTPClientFactory):
    def setURL(self, url):
        HTTPClientFactory.setURL(self, url)
        self.path = url


def receive():
    global _ReceivingLoop
    lg.out(6, 'http_node.receive')

    if _ReceivingLoop is not None:
        return _ReceivingLoop

    def success(src, idurl, host, port, conn):
        global _LastPingTimeDict
        global _ConnectionsDict

        if len(src) == 0:
            increase_receiving_delay(idurl)

        else:
            parts = src.splitlines()
            lg.out(14, 'http_node.receive.success %d bytes in %d parts from %s (%s:%s)' % (len(src), len(parts), idurl, host, port))
            for part64 in parts:
                try:
                    part = base64.b64decode(part64.strip())
                except:
                    lg.out(14, 'http_node.receive.success ERROR in base64.b64decode()')
                    #                    lg.out(1, 'len(part64)=%d' % len(part64))
                    #                    lg.out(1, 'len(part64.strip())=%d' % len(part64.strip()))
                    #                    lg.out(1, 'part64=[%s]' % part64)
                    decrease_receiving_delay(idurl)
                    continue
#                fd, filename = tempfile.mkstemp(".dhn-http-in")
                fd, filename = tmpfile.make('http-in', extension='.http')
                os.write(fd, part)
                os.close(fd)
                decrease_receiving_delay(idurl)
                transport_control.receiveStatusReport(
                    filename,
                    'finished',
                    'http',
                    host + ':' + port,
                )
            transport_control.log('http', 'finish connection with %s:%s ' % (host, port))

        conn.disconnect()
        _ConnectionsDict.pop(idurl, None)

    def fail(x, idurl, host, port, conn):
        global _LastPingTimeDict
        global _ConnectionsDict
        increase_receiving_delay(idurl)
        conn.disconnect()
        _ConnectionsDict.pop(idurl, None)

    def ping(idurl, host, port):
        lg.out(14, 'http_node.receive.ping     %s (%s:%s)' % (idurl, host, port))
        url = 'http://' + str(host) + ':' + str(port)

        if dhnnet.proxy_is_on():
            f = TransportHTTPProxyClientFactory(
                url,
                method='POST',
                headers={
                    'User-Agent': 'BitDust transport_http',
                    'idurl': misc.getLocalID(),
                },
            )
            conn = reactor.connectTCP(dhnnet.get_proxy_host(), int(dhnnet.get_proxy_port()), f)
        else:
            f = TransportHTTPClientFactory(
                url,
                method='POST',
                headers={
                    'User-Agent': 'BitDust transport_http',
                    'idurl': misc.getLocalID(),
                },
            )
            conn = reactor.connectTCP(host, int(port), f)


#        f = HTTPClientFactory(url, method='POST', headers={
#            'User-Agent': 'BitDust transport_http',
#            'idurl': misc.getLocalID(), } )
#        conn = reactor.connectTCP(host, int(port), f)

        f.deferred.addCallback(success, idurl, host, port, conn)
        f.deferred.addErrback(fail, idurl, host, port, conn)
        return conn

    def loop():
        global _ReceivingLoop
        global _Contacts
        global _ToIncreaseDelay
        global _LastPingTimeDict
        global _PingDelayDict
        global _ConnectionsDict
        global _CurrentDelay
        _ReceivingLoop = reactor.callLater(1, loop)
        _CurrentDelay = settings.getHTTPDelay()

        for idurl, hostport in _Contacts.items():
            if idurl in _ConnectionsDict:
                continue

            lasttm = _LastPingTimeDict.get(idurl, 0)
            delay = _PingDelayDict.get(idurl, _CurrentDelay)
            dt = time.time() - lasttm

            if dt < delay:
                continue

            _ConnectionsDict[idurl] = ping(idurl, hostport[0], hostport[1])
            _LastPingTimeDict[idurl] = time.time()

        return _ReceivingLoop

    return loop()


def decrease_receiving_delay(idurl):
    global _PingDelayDict
    global _CurrentDelay
    lg.out(14, 'http_node.decrease_receiving_delay ' + idurl)
    _PingDelayDict[idurl] = _CurrentDelay


def increase_receiving_delay(idurl):
    global _PingDelayDict
    global _CurrentDelay
    if idurl not in _PingDelayDict:
        _PingDelayDict[idurl] = _CurrentDelay
    d = _PingDelayDict[idurl]
    if d < settings.DefaultSendTimeOutHTTP()/2:
        lg.out(14, 'http_node.increase_receiving_delay   %s for %s' % (str(d), idurl))
        _PingDelayDict[idurl] *= 2


def stop_receiving():
    global _ReceivingLoop
    lg.out(6, 'http_node.stop_receiving')
    if _ReceivingLoop is None:
        lg.out(8, 'http_node.stop_receiving _ReceivingLoop is None')
        return
    _ReceivingLoop.cancel()
    del _ReceivingLoop
    _ReceivingLoop = None


def add_contact(idurl):
    global _Contacts
    global _PingDelayDict
    global _CurrentDelay
    lg.out(14, 'http_node.add_contact want to add %s' % idurl)
    ident = contacts.getContact(idurl)
    if ident is None:
        lg.out(6, 'http_node.add_contact WARNING %s not in contacts' % idurl)
        return
    http_contact = ident.getProtoContact('http')
    if http_contact is None:
        lg.out(12, 'http_node.add_contact %s have no http contact. skip.' % idurl)
        return
    proto, host, port, filename = nameurl.UrlParse(http_contact)
    _Contacts[idurl] = (host, port)
    _PingDelayDict[idurl] = _CurrentDelay
    lg.out(10, 'http_node.add_contact %s on %s:%s' % (idurl, host, port))


def update_contacts():
    lg.out(10, 'http_node.update_contacts ')
    global _Contacts
    _Contacts.clear()

    def update_contact(x, idurl):
        add_contact(idurl)

    def failed(x, idurl):
        lg.out(10, 'http_node.update_contacts.failed   NETERROR ' + idurl)

    contacts_list = contacts.getContactIDs()
    contacts_list.append(settings.CentralID())
    contacts_list.append(settings.MoneyServerID())
    contacts_list.append(settings.MarketServerID())
    for idurl in contacts_list:
        lg.out(10, 'http_node.update_contacts want ' + idurl)
        if idurl == misc.getLocalID():
            continue
        ident = contacts.getContact(idurl)
        if ident is None:
            d = identitycache.immediatelyCaching(idurl)
            d.addCallback(update_contact, idurl)
            d.addErrback(failed, idurl)
            continue

        update_contact('', idurl)


def contacts_changed_callback(oldlist, newlist):
    update_contacts()


def init():
    lg.out(4, 'http_node.init')
    contacts.SetContactsChangedCallback(contacts_changed_callback)
    update_contacts()


def shutdown():
    global _ReceivingLoop
    global _ServerListener
    if _ReceivingLoop is not None:
        _ReceivingLoop.cancel()
        _ReceivingLoop = None
    if _ServerListener is not None:
        _ServerListener.stopListening()
        _ServerListener = None


#-------------------------------------------------------------------------------


def usage():
    print('''usage:
http_node.py send [server_port] [to idurl] [filename]
http_node.py receive
''')


def main():
    if sys.argv.count('receive'):
        ##        log.startLogging(sys.stdout)
        settings.init()
        settings.update_proxy_settings()
        contacts.init()
        init()
        receive()
        reactor.run()
    elif sys.argv.count('send'):
        ##        log.startLogging(sys.stdout)
        settings.init()
        settings.update_proxy_settings()
        contacts.init()
        init()
        start_http_server(int(sys.argv[2]))
        send(sys.argv[3], sys.argv[4])
        reactor.run()
    else:
        usage()


if __name__ == '__main__':
    main()



#!/usr/bin/python
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (__init__.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
`BitDust <https://bitdust.io>`_ is a peer to peer online backup utility.

This is a distributed network for backup data storage.
Each participant of the network provides a portion of his hard drive for other users.
In exchange, he is able to store his data on other peers.

The redundancy in backup makes it so if someone loses your data,
you can rebuild what was lost and give it to someone else to hold.
And all of this happens without you having to do a thing - the software keeps your data in safe.

All your data is encrypted before it leaves your computer with a Private key your computer generates.
No one else can read your data, even BitDust !
Recover data is only one way - download the necessary pieces from computers of other peers
and decrypt them with your private key.

`BitDust <https://bitdust.io>`_ is written in `Python <http://python.org/>`_
using pure `Twisted framework <http://twistedmatrix.com/>`_.
"""



#!/usr/bin/python
# friends.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (friends.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
"""
.. module:: friends

"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 4

#------------------------------------------------------------------------------

import sys

from twisted.internet.defer import Deferred

#------------------------------------------------------------------------------

if __name__ == '__main__':
    import os.path as _p
    sys.path.insert(0, _p.abspath(_p.join(_p.dirname(_p.abspath(sys.argv[0])), '..')))

#------------------------------------------------------------------------------

from bitdust.lib import strng

from bitdust.userid import global_id
from bitdust.userid import id_url

#------------------------------------------------------------------------------

# def do_add(idurl, result_defer, alias):
#     added = False
#     if not contactsdb.is_correspondent(idurl):
#         contactsdb.add_correspondent(idurl, alias)
#         contactsdb.save_correspondents()
#         added = True
#         events.send('friend-added', data=dict(
#             idurl=idurl,
#             global_id=global_id.idurl2glob(idurl),
#             alias=alias,
#         ))
#     d = online_status.handshake(idurl, channel='friend_add', keep_alive=True)
#     if share_person_key:
#         from bitdust.access import key_ring
#         from bitdust.crypt import my_keys
#         my_person_key_id = my_id.getGlobalID(key_alias='person')
#         if my_keys.is_key_registered(my_person_key_id):
#             d.addCallback(lambda *args: [
#                 key_ring.share_key(
#                     key_id=my_person_key_id,
#                     trusted_idurl=idurl,
#                     include_private=False,
#                     include_signature=True,
#                     timeout=15,
#                 ),
#                 result_defer.callback(),
#             ])
#
#     if _Debug:
#         d.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, method='api.friend_add')
#     d.addErrback(result_defer)
#     if added:
#         return OK(message='new friend has been added', api_method='friend_add')
#     return OK(message='this friend has been already added', api_method='friend_add')


def add_friend(trusted_user_id, alias='', share_person_key=True):
    idurl = strng.to_text(trusted_user_id)
    if global_id.IsValidGlobalUser(trusted_user_id):
        idurl = global_id.GlobalUserToIDURL(trusted_user_id, as_field=False)
    idurl = id_url.field(idurl)
    if not idurl:
        return None

    ret = Deferred()


#     if id_url.is_cached(idurl):
#         _add(idurl, ret)
#         return ret
#
#     d = identitycache.immediatelyCaching(idurl)
#     d.addErrback(lambda *args: ret.callback(ERROR('failed caching user identity')))
#     d.addCallback(lambda *args: _add(idurl, ret))
#     return ret



#!/usr/bin/env python
# file_down.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (file_down.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""

.. raw:: html

    <a href="https://bitdust.io/automats/customer/file_down.png" target="_blank">
    <img src="https://bitdust.io/automats/customer/file_down.png" style="max-width:100%;">
    </a>


.. module:: file_down
.. role:: red


BitDust file_down() Automat

EVENTS:
    * :red:`fail-received`
    * :red:`file-already-exists`
    * :red:`init`
    * :red:`request-failed`
    * :red:`retrieve-sent`
    * :red:`start`
    * :red:`stop`
    * :red:`valid-data-received`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 16

#------------------------------------------------------------------------------

import sys
import time

#------------------------------------------------------------------------------

try:
    from twisted.internet import reactor  # @UnresolvedImport
except:
    sys.exit('Error initializing twisted.internet.reactor in file_down.py')

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.automats import automat

from bitdust.lib import utime
from bitdust.lib import packetid
from bitdust.lib import nameurl

from bitdust.userid import global_id

from bitdust.main import settings

from bitdust.p2p import p2p_service
from bitdust.p2p import commands

from bitdust.transport import packet_out

from bitdust.stream import io_throttle

#------------------------------------------------------------------------------


class FileDown(automat.Automat):
    """
    This class implements all the functionality of ``file_down()`` state machine.
    """
    def __init__(self, parent, callOnReceived, creatorID, packetID, ownerID, remoteID, debug_level=_DebugLevel, log_events=_Debug, log_transitions=_Debug, publish_events=False, **kwargs):
        """
        Builds `file_down()` state machine.
        """
        self.parent = parent
        self.callOnReceived = []
        self.callOnReceived.append(callOnReceived)
        self.creatorID = creatorID
        self.packetID = global_id.CanonicalID(packetID)
        parts = global_id.NormalizeGlobalID(packetID)
        self.customerID = parts['customer']
        self.remotePath = parts['path']
        self.customerIDURL = parts['idurl']
        customerGlobalID, remotePath, versionName, fileName = packetid.SplitVersionFilename(packetID)
        self.backupID = packetid.MakeBackupID(customerGlobalID, remotePath, versionName)
        self.fileName = fileName
        self.ownerID = ownerID
        self.remoteID = remoteID
        self.requestTime = None
        self.fileReceivedTime = None
        self.requestTimeout = max(30, 2*int(settings.getBackupBlockSize()/settings.SendingSpeedLimit()))
        self.result = ''
        self.created = utime.utcnow_to_sec1970()
        super(FileDown, self).__init__(
            name='file_down_%s_%s/%s/%s' % (nameurl.GetName(self.remoteID), remotePath, versionName, fileName), state='AT_STARTUP', debug_level=debug_level, log_events=log_events, log_transitions=log_transitions,
            publish_events=publish_events, **kwargs
        )

    def A(self, event, *args, **kwargs):
        """
        The state machine code, generated using `visio2python <http://bitdust.io/visio2python/>`_ tool.
        """
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'init':
                self.state = 'IN_QUEUE'
                self.doInit(*args, **kwargs)
                self.doQueueAppend(*args, **kwargs)
        #---IN_QUEUE---
        elif self.state == 'IN_QUEUE':
            if event == 'file-already-exists':
                self.state = 'EXIST'
                self.doQueueRemove(*args, **kwargs)
                self.doReportExist(*args, **kwargs)
                self.doQueueNext(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'start':
                self.state = 'STARTED'
                self.doSendRetreive(*args, **kwargs)
            elif event == 'stop':
                self.state = 'STOPPED'
                self.doQueueRemove(*args, **kwargs)
                self.doReportStopped(*args, **kwargs)
                self.doQueueNext(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
        #---STARTED---
        elif self.state == 'STARTED':
            if event == 'stop':
                self.state = 'STOPPED'
                self.doCancelPackets(*args, **kwargs)
                self.doQueueRemove(*args, **kwargs)
                self.doReportStopped(*args, **kwargs)
                self.doQueueNext(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'request-failed':
                self.state = 'FAILED'
                self.doQueueRemove(*args, **kwargs)
                self.doReportFailed(event, *args, **kwargs)
                self.doQueueNext(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'retrieve-sent':
                self.state = 'REQUESTED'
        #---REQUESTED---
        elif self.state == 'REQUESTED':
            if event == 'valid-data-received':
                self.state = 'RECEIVED'
                self.doQueueRemove(*args, **kwargs)
                self.doReportReceived(*args, **kwargs)
                self.doQueueNext(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'fail-received':
                self.state = 'FAILED'
                self.doQueueRemove(*args, **kwargs)
                self.doReportFailed(event, *args, **kwargs)
                self.doQueueNext(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'stop':
                self.state = 'STOPPED'
                self.doCancelPackets(*args, **kwargs)
                self.doQueueRemove(*args, **kwargs)
                self.doReportStopped(*args, **kwargs)
                self.doQueueNext(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
        #---EXIST---
        elif self.state == 'EXIST':
            pass
        #---FAILED---
        elif self.state == 'FAILED':
            pass
        #---STOPPED---
        elif self.state == 'STOPPED':
            pass
        #---RECEIVED---
        elif self.state == 'RECEIVED':
            pass
        return None

    def doInit(self, *args, **kwargs):
        """
        Action method.
        """
        io_throttle.PacketReport('request', self.remoteID, self.packetID, 'init')

    def doQueueAppend(self, *args, **kwargs):
        """
        Action method.
        """
        if self.packetID in self.parent.fileRequestQueue:
            raise Exception('file %r already in downloading queue for %r' % (self.packetID, self.remoteID))
        if self.packetID in self.parent.fileRequestDict:
            raise Exception('file %r already in downloading dict for %r' % (self.packetID, self.remoteID))
        self.parent.fileRequestDict[self.packetID] = self
        self.parent.fileRequestQueue.append(self.packetID)

    def doQueueRemove(self, *args, **kwargs):
        """
        Action method.
        """
        if self.packetID not in self.parent.fileRequestDict:
            raise Exception('file %r not found in downloading dict for %r' % (self.packetID, self.remoteID))
        if self.packetID not in self.parent.fileRequestQueue:
            raise Exception('file %r not found in downloading queue for %r' % (self.packetID, self.remoteID))
        self.parent.fileRequestQueue.remove(self.packetID)
        del self.parent.fileRequestDict[self.packetID]

    def doSendRetreive(self, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.args(_DebugLevel, packetID=self.packetID, remoteID=self.remoteID)
        p2p_service.SendRetreive(
            self.ownerID,
            self.parent.creatorID,
            self.packetID,
            self.remoteID,
            callbacks={
                commands.Data(): self.parent.OnDataReceived,
                commands.Fail(): self.parent.OnDataReceived,
                # None: lambda pkt_out: self.OnDataReceived(fileRequest.packetID, 'timeout'),  # timeout
            },
            # response_timeout=10,
        )
        self.requestTime = time.time()

    def doCancelPackets(self, *args, **kwargs):
        """
        Action method.
        """
        packetsToCancel = packet_out.search_by_packet_id(self.packetID)
        for pkt_out in packetsToCancel:
            if pkt_out.outpacket.Command == commands.Retrieve():
                if _Debug:
                    lg.dbg(_DebugLevel, 'sending "cancel" to %s addressed to %s because downloading cancelled' % (pkt_out, pkt_out.remote_idurl))
                pkt_out.automat('cancel')

    def doQueueNext(self, *args, **kwargs):
        """
        Action method.
        """
        self.parent.DoRequest()

    def doReportExist(self, *args, **kwargs):
        """
        Action method.
        """
        for callBack in self.callOnReceived:
            reactor.callLater(0, callBack, self.packetID, 'exist')  # @UndefinedVariable

    def doReportStopped(self, *args, **kwargs):
        """
        Action method.
        """
        for callBack in self.callOnReceived:
            reactor.callLater(0, callBack, self.packetID, 'cancelled')  # @UndefinedVariable

    def doReportReceived(self, *args, **kwargs):
        """
        Action method.
        """
        self.fileReceivedTime = time.time()
        for callBack in self.callOnReceived:
            reactor.callLater(0, callBack, args[0], 'received')  # @UndefinedVariable

    def doReportFailed(self, event, *args, **kwargs):
        """
        Action method.
        """
        if event == 'fail-received':
            for callBack in self.callOnReceived:
                reactor.callLater(0, callBack, args[0], 'failed')  # @UndefinedVariable
        else:
            for callBack in self.callOnReceived:
                reactor.callLater(0, callBack, self.packetID, 'failed')  # @UndefinedVariable

    def doDestroyMe(self, *args, **kwargs):
        """
        Remove all references to the state machine object to destroy it.
        """
        io_throttle.PacketReport('request', self.remoteID, self.packetID, self.result)
        self.parent = None
        self.callOnReceived = None
        self.creatorID = None
        self.packetID = None
        self.customerID = None
        self.remotePath = None
        self.customerIDURL = None
        self.backupID = None
        self.fileName = None
        self.ownerID = None
        self.remoteID = None
        self.requestTime = None
        self.fileReceivedTime = None
        self.requestTimeout = None
        self.result = None
        self.created = None
        self.destroy()



#!/usr/bin/python
# service_contract_chain.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (service_contract_chain.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: service_contract_chain
"""

from __future__ import absolute_import
from bitdust.services.local_service import LocalService


def create_service():
    return ContractChainService()


class ContractChainService(LocalService):

    service_name = 'service_contract_chain'
    config_path = 'services/contract-chain/enabled'

    def dependent_on(self):
        return [
            'service_nodes_lookup',
        ]

    def installed(self):
        # TODO: to be continue...
        return False

    def start(self):
        from twisted.internet.defer import Deferred
        from coins import contract_chain_consumer
        self.starting_deferred = Deferred()
        contract_chain_consumer.A('init')
        contract_chain_consumer.A().addStateChangedCallback(self._on_contract_chain_state_changed)
        contract_chain_consumer.A('start')
        return self.starting_deferred

    def stop(self):
        from coins import contract_chain_consumer
        contract_chain_consumer.A().removeStateChangedCallback(self._on_contract_chain_state_changed)
        contract_chain_consumer.A('stop')
        contract_chain_consumer.A('shutdown')
        return True

    def health_check(self):
        from coins import contract_chain_consumer
        return contract_chain_consumer.A().state in [
            'CONNECTED',
        ]

    def _on_contract_chain_state_changed(self, oldstate, newstate, event_string, *args, **kwargs):
        if self.starting_deferred:
            if newstate in [
                'CONNECTED',
            ] and oldstate not in [
                'AT_STARTUP',
            ]:
                self.starting_deferred.callback(True)
                self.starting_deferred = None
            elif newstate in [
                'DISCONNECTED',
            ] and oldstate not in [
                'AT_STARTUP',
            ]:
                self.starting_deferred.errback(Exception(newstate))
                self.starting_deferred = None



#!/usr/bin/python
# tmpfile.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (tmpfile.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: tmpfile.

Keep track of temporary files created in the program. The temp folder is
placed in the BitDust data directory. All files are divided into several
sub folders.
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

import os
import tempfile
import time

from twisted.internet import task  # @UnresolvedImport

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.system import bpio

#------------------------------------------------------------------------------

_TempDirPath = None
_FilesDict = {}
_CollectorTask = None
_SubDirs = {
    'outbox': 60*10,  # hold onto outbox files 10 minutes
    'tcp-in': 60*10,  # 10 minutes for incoming tcp files
    'udp-in': 60*10,  # 10 minutes for incoming udp files
    'proxy-in': 60*10,  # 10 minutes for incoming proxy files
    'proxy-out': 60*10,  # 10 minutes for outgoing proxy files
    'propagate': 60*10,  # propagate happens often enough,
    'backup': 60*10,  # 10 minutes for backup files
    'restore': 0,  # never remove files during restore process, they must be cleaned afterwards
    'raid': 60*10,  # 10 minutes for backup files
    'idsrv': 60,  # 1 minute for incoming xml identity files
    'error': 60*60*24*30,  # store errors for one month
    'all': 0,  # other files. do not know when to remove
    # they can be located even in another location
    # use register(name, filename)
}

#------------------------------------------------------------------------------


def init(temp_dir_path=''):
    """
    Must be called before all other things here.

    - check existence and access mode of temp folder
    - creates a needed sub folders
    - call ``startup_clean()``
    - starts collector task to call method ``collect()`` every 5 minutes
    """
    if _Debug:
        lg.out(_DebugLevel, 'tmpfile.init')
    global _TempDirPath
    global _SubDirs
    global _FilesDict
    global _CollectorTask

    if _TempDirPath is None:
        if temp_dir_path != '':
            _TempDirPath = temp_dir_path
        else:
            os_temp_dir = tempfile.gettempdir()
            temp_dir = os.path.join(os_temp_dir, 'bitdust')

            if not os.path.exists(temp_dir):
                try:
                    os.mkdir(temp_dir)
                except:
                    lg.exc()
                    temp_dir = os_temp_dir

            if not os.access(temp_dir, os.W_OK):
                lg.out(2, 'tmpfile.init ERROR no write permissions to ' + temp_dir)
                temp_dir = os_temp_dir

            _TempDirPath = temp_dir
        lg.out(6, 'tmpfile.init  _TempDirPath=' + _TempDirPath)

    for name in _SubDirs.keys():
        if not os.path.exists(subdir(name)):
            try:
                os.makedirs(subdir(name))
            except:
                lg.out(2, 'tmpfile.init ERROR can not create ' + subdir(name))
                lg.exc()

    for name in _SubDirs.keys():
        if name not in _FilesDict:
            _FilesDict[name] = {}

    startup_clean()

    if _CollectorTask is None:
        _CollectorTask = task.LoopingCall(collect)
        _CollectorTask.start(60*5)


def shutdown():
    """
    Do not need to remove any files here, just stop the collector task.
    """
    if _Debug:
        lg.out(_DebugLevel, 'tmpfile.shutdown')
    global _CollectorTask
    if _CollectorTask is not None:
        _CollectorTask.stop()
        del _CollectorTask
        _CollectorTask = None


def subdir(name):
    """
    Return a path to given sub folder.
    """
    global _TempDirPath
    if _TempDirPath is None:
        init()
    return os.path.join(_TempDirPath, name)


def register(filepath):
    """
    You can create a temp file in another place and call this method to be able
    to hadle this file later.
    """
    global _FilesDict
    subdir, _ = os.path.split(filepath)
    name = os.path.basename(subdir)
    if name not in list(_FilesDict.keys()):
        name = 'all'
    _FilesDict[name][filepath] = time.time()


def make(name, extension='', prefix='', close_fd=False):
    """
    Make a new binary file under sub folder ``name`` and return a tuple of it's file
    descriptor and path.

    .. warning::    Remember you need to close the file descriptor by your own.
    The ``tmpfile`` module will remove it later - do not worry.
    This is a job for our collector.
    However if you will keep the file opened for awhile
    it should print a ```WARNING``` in logs because will fail to delete it.
    """
    global _TempDirPath
    global _FilesDict
    if _TempDirPath is None:
        init()
    if name not in list(_FilesDict.keys()):
        name = 'all'
    try:
        fd, filename = tempfile.mkstemp(extension, prefix, subdir(name))
        _FilesDict[name][filename] = time.time()
    except:
        lg.out(1, 'tmpfile.make ERROR creating file in sub folder ' + name)
        lg.exc()
        return None, ''
    if close_fd:
        os.close(fd)
    if _Debug:
        lg.out(_DebugLevel, 'tmpfile.make ' + filename)
    return fd, filename


def make_dir(name, extension='', prefix=''):
    global _TempDirPath
    global _FilesDict
    if _TempDirPath is None:
        init()
    if name not in list(_FilesDict.keys()):
        name = 'all'
    try:
        dirname = tempfile.mkdtemp(extension, prefix, subdir(name))
        _FilesDict[name][dirname] = time.time()
    except:
        lg.out(1, 'tmpfile.make_dir ERROR creating folder in ' + name)
        lg.exc()
        return None
    if _Debug:
        lg.out(_DebugLevel, 'tmpfile.make_dir ' + dirname)
    return dirname


def erase(name, filename, why='no reason'):
    """
    However you can remove not needed file immediately, this is a good way
    also.

    But outside of this module you better use method ``throw_out``.
    """
    global _FilesDict
    if name in list(_FilesDict.keys()):
        try:
            _FilesDict[name].pop(filename)
        except:
            pass
            # lg.warn('we do not know about item [%s] in sub folder %s, we tried because %s' % (filename, name, why))
    else:
        lg.warn('we do not know sub folder: %s, we tried because %s' % (name, why))

    if not os.path.exists(filename):
        lg.warn('[%s] not exist' % filename)
        return

    if os.path.isfile(filename):
        if not os.access(filename, os.W_OK):
            lg.warn('[%s] no write permissions' % filename)
            return
        try:
            os.remove(filename)
            if _Debug:
                lg.out(_DebugLevel, 'tmpfile.erase [%s] : "%s"' % (filename, why))
        except:
            lg.out(2, 'tmpfile.erase ERROR can not remove [%s], we tried because %s' % (filename, why))

    elif os.path.isdir(filename):
        bpio.rmdir_recursive(filename, ignore_errors=True)
        if _Debug:
            lg.out(_DebugLevel, 'tmpfile.erase recursive [%s] : "%s"' % (filename, why))

    else:
        raise Exception('path [%s] not exist' % filename)


def throw_out(filepath, why='dont know'):
    """
    A more smart way to remove not needed temporary file, accept a full
    ``filepath``.
    """
    global _FilesDict
    global _SubDirs
    subdir, _ = os.path.split(filepath)
    name = os.path.basename(subdir)
    erase(name, filepath, why)


def collect():
    """
    Removes old temporary files.
    """
    if _Debug:
        lg.out(_DebugLevel - 4, 'tmpfile.collect')
    global _FilesDict
    global _SubDirs
    erase_list = []
    for name in _FilesDict.keys():
        # for data and parity files we have special rules
        # we do not want to remove Data or Parity here.
        # backup_monitor should take care of this.
        if name == 'data-par':
            continue

        else:
            # how long we want to keep the file?
            lifetime = _SubDirs.get(name, 0)
            # if this is not set - keep forever
            if lifetime == 0:
                continue
            for filename, filetime in _FilesDict[name].items():
                # if file is too old - remove it
                if time.time() - filetime > lifetime:
                    erase_list.append((name, filename))

    for name, filename in erase_list:
        erase(name, filename, 'collected')

    if _Debug:
        lg.out(_DebugLevel - 4, 'tmpfile.collect %d files erased' % len(erase_list))

    del erase_list


def startup_clean():
    """
    At startup we want to scan all sub folders and remove the old files.

    We will get creation time with built-in ``os.stat`` method.
    """
    global _TempDirPath
    global _SubDirs
    if _Debug:
        lg.out(_DebugLevel - 4, 'tmpfile.startup_clean in %s' % _TempDirPath)
    if _TempDirPath is None:
        return
    counter = 0
    limit_counts = 200
    for name in os.listdir(_TempDirPath):
        # we want to scan only our folders
        # do not want to be responsible of other files
        if name not in list(_SubDirs.keys()):
            continue

        # for data and parity files we have special rules
        # we do not want to remove Data or Parity here.
        # backup_monitor should take care of this.
        if name == 'data-par':
            pass

        else:
            lifetime = _SubDirs.get(name, 0)
            if lifetime == 0:
                continue
            for filename in os.listdir(subdir(name)):
                filepath = os.path.join(subdir(name), filename)
                if os.path.isfile(filepath):
                    filetime = os.stat(filepath).st_ctime
                    if time.time() - filetime > lifetime:
                        erase(name, filepath, 'startup cleaning')
                        counter += 1
                        if counter > limit_counts:
                            break
                elif os.path.isdir(filepath):
                    erase(name, filepath, 'startup cleaning')
                    counter += 1
                    if counter > limit_counts:
                        break
                else:
                    raise Exception('path %s not exist' % filepath)


#------------------------------------------------------------------------------

if __name__ == '__main__':
    lg.set_debug_level(18)
    init()
    fd, filename = make('raid')
    os.write(fd, 'TEST FILE')
    os.close(fd)
    from twisted.internet import reactor  # @UnresolvedImport
    reactor.run()  # @UndefinedVariable



# optiexplorer.py v 0.32 to be used with Python3.5 or better
# Copyright Hclivess, Maccaspacca, vv181 2017
# Copyright Maccaspacca 2018
# for license see LICENSE file
# ..

from tornado.wsgi import WSGIContainer
from tornado.httpserver import HTTPServer
from tornado.ioloop import IOLoop

import sqlite3, time, essentials
from flask import Flask, render_template

app = Flask(__name__)

key, public_key_readable, private_key_readable, _, _, public_key_hashed, address, keyfile = essentials.keys_load('privkey.der', 'pubkey.der')

# load config

try:

    lines = [line.rstrip('\n') for line in open('pool.txt')]
    for line in lines:
        try:
            if 'm_timeout=' in line:
                m_timeout = int(line.split('=')[1])
        except Exception as e:
            m_timeout = 5

except Exception as e:
    m_timeout = 5

# load config

#@app.route('/static/<filename>')
# def server_static(filename):
# return static_file(filename, root='static/')


@app.route('/')
# def hello():
def main():

    conn = sqlite3.connect('static/ledger.db', timeout=60.0)
    conn.text_factory = str
    c = conn.cursor()

    shares = sqlite3.connect('shares.db')
    shares.text_factory = str
    s = shares.cursor()

    oldies = sqlite3.connect('archive.db')
    oldies.text_factory = str
    o = oldies.cursor()

    addresses = []
    for row in s.execute('SELECT * FROM shares'):
        shares_address = row[0]
        shares_value = row[1]
        shares_timestamp = row[2]

        if shares_address not in addresses:
            addresses.append(shares_address)

    total_hash = 0
    worker_count = 0
    output_shares = []
    output_timestamps = []

    data_addres = []
    data_shares = []
    data_mrate = []
    data_mname = []
    data_wcount = []

    for x in addresses:
        s.execute('SELECT sum(shares) FROM shares WHERE address = ? AND paid != 1', (x, ))
        shares_sum = s.fetchone()[0]
        if shares_sum == None:
            shares_sum = 0
            continue
        output_shares.append(shares_sum)

        s.execute('SELECT timestamp FROM shares WHERE address = ? ORDER BY timestamp ASC LIMIT 1', (x, ))
        shares_timestamp = s.fetchone()[0]
        output_timestamps.append(float(shares_timestamp))

        s.execute('SELECT * FROM shares WHERE address = ? ORDER BY timestamp DESC LIMIT 1', (x, ))
        shares_last = s.fetchone()
        #mrate = shares_last[4]
        mname = shares_last[7]  # last worker

        s.execute('SELECT DISTINCT name FROM shares WHERE address = ?', (x, ))
        shares_names = s.fetchall()

        nrate = []
        ncount = []
        for n in shares_names:
            s.execute('SELECT * FROM shares WHERE address = ? AND name = ? ORDER BY timestamp DESC LIMIT 1', (x, n[0]))
            names_last = s.fetchone()
            t1 = time.time()
            t2 = float(names_last[2])
            t3 = (t1 - t2)/60
            if t3 < m_timeout:
                nrate.append(int(names_last[4]))
                ncount.append(int(names_last[6]))
            else:
                nrate.append(0)
                ncount.append(0)

        mrate = sum(nrate)  # hashrate of address
        wcount = sum(ncount)  # worker count
        total_hash = total_hash + mrate
        worker_count = worker_count + wcount

        if t3 < 30:

            data_addres.append(x)
            data_shares.append(shares_sum)
            data_mrate.append(str(mrate))
            data_mname.append(mname)
            data_wcount.append(str(wcount))

    try:
        shares_total = sum(output_shares)
    except:
        shares_total = 0

    try:
        block_threshold = min(output_timestamps)
    except:
        block_threshold = time.time()

    reward_list = []
    data_block = []
    data_reward = []
    data_tShares = []
    data_rewardps = []
    data_tReward = []
    data_tHash = []
    data_twcount = []

    for row in c.execute('SELECT * FROM transactions WHERE address = ? AND CAST(timestamp AS INTEGER) >= ? AND reward != 0', (address, ) + (block_threshold, )):
        data_block.append(row[0])
        data_reward.append(row[9])
        reward_list.append(float(row[9]))

    if data_block == [] and data_reward == []:
        data_block.append(0)
        data_reward.append(0)

    reward_total = sum(reward_list)

    try:
        reward_per_share = reward_total/shares_total
    except:
        reward_per_share = 0

    data_tShares.append(shares_total)
    data_rewardps.append(reward_per_share)
    data_tReward.append(reward_total)
    data_tHash.append(format('%.2f' % (total_hash/1000)))
    data_twcount.append(worker_count)

    # payout view
    data_pendingaddress = []
    data_pendingreward = []
    if reward_total > 0:

        for x, y in zip(addresses, output_shares):

            try:
                claim = y*reward_per_share
            except:
                claim = 0

            data_pendingaddress.append(x)
            data_pendingreward.append(format('%.8f' % (claim)))

    data_paddress = []
    data_bismuthreward = []
    data_blockheight = []
    data_ptime = []

    for row in c.execute('SELECT * FROM transactions WHERE address = ? and openfield = ? ORDER BY timestamp DESC LIMIT 80', (address, ) + ('pool', )):
        data_paddress.append(row[3])
        data_bismuthreward.append(row[4])
        data_blockheight.append(row[0])
        data_ptime.append(format(time.strftime('%Y/%m/%d,%H:%M:%S', time.gmtime(float(row[1])))))

    conn.close()
    shares.close()
    oldies.close()

    return render_template(
        'index.html',
        recentminers=zip(data_addres, data_shares, data_mrate, data_mname, data_wcount),
        bpstats=zip(data_block, data_reward, data_tShares, data_rewardps, data_tReward, data_tHash, data_twcount),
        payouts=zip(data_addres, data_bismuthreward, data_blockheight, data_ptime),
        payoutsfees=zip(data_pendingaddress, data_pendingreward),
    )


if __name__ == '__main__':
    #app.run(host='0.0.0.0', port=9080, debug=True)
    http_server = HTTPServer(WSGIContainer(app))
    http_server.listen(9080)
    IOLoop.instance().start()



#!/usr/bin/python
# run_upnpc.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (run_upnpc.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: run_upnpc
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from __future__ import print_function

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 6

#------------------------------------------------------------------------------

import os
import sys
import re
import random

#------------------------------------------------------------------------------

if __name__ == '__main__':
    import os.path as _p
    sys.path.append(_p.join(_p.dirname(_p.abspath(sys.argv[0])), '..'))

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.system import bpio

from bitdust.main import settings

#------------------------------------------------------------------------------

_CurrentProcess = None
_MyPortMapping = {}
_LastUpdateResultDict = {}

#-------------------------------------------------------------------------------


def init():
    global _MyPortMapping
    if _Debug:
        lg.out(_DebugLevel, 'run_upnpc.init ')


def shutdown():
    global _CurrentProcess
    if _Debug:
        lg.out(_DebugLevel, 'run_upnpc.shutdown')
    if _CurrentProcess is not None:
        if _Debug:
            lg.out(_DebugLevel, '    going to kill _CurrentProcess=%s' % _CurrentProcess)
        try:
            _CurrentProcess.kill()
            if _Debug:
                lg.out(_DebugLevel, '    killed')
        except:
            lg.exc()
    else:
        if _Debug:
            lg.out(_DebugLevel, '    not started')


#------------------------------------------------------------------------------


def execute_in_shell(cmdargs, base_dir=None):
    global _CurrentProcess
    from bitdust.system import nonblocking
    import subprocess
    if _Debug:
        lg.out(_DebugLevel, 'run_upnpc.execute_in_shell: "%s"' % (' '.join(cmdargs)))
    in_shell = True
    if bpio.Mac():
        in_shell = False
    _CurrentProcess = nonblocking.Popen(
        cmdargs,
        shell=in_shell,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )
    out_data = _CurrentProcess.communicate()[0]
    returncode = _CurrentProcess.returncode
    if _Debug:
        lg.out(_DebugLevel, 'run_upnpc.execute_in_shell returned: %s and %d bytes output' % (returncode, len(out_data)))
    if returncode > 0:
        if _Debug:
            lg.out(_DebugLevel, '\n' + out_data)
    return (out_data, returncode)  # _CurrentProcess


#------------------------------------------------------------------------------

# Windows: executable file "upnpc.exe" must be in same folder or system sub folder
# Ubuntu: miniupnpc must be installed, https://launchpad.net/ubuntu/+source/miniupnpc


def run(args_list, base_dir=None, callback=None):
    # TODO: Currently disabled
    return None

    global _CurrentProcess
    if _CurrentProcess is not None:
        lg.warn('only one process at once')
        return None

    if bpio.Windows():
        cmdargs = [
            'upnpc-static.exe',
        ]
    elif bpio.Linux():
        cmdargs = [
            'upnpc',
        ]
    elif bpio.Mac():
        cmdargs = [
            'upnpc',
        ]
    else:
        return None

    if bpio.Windows():
        # if we run windows binaries - upnpc-static.exe can be in the system sub folder
        if not os.path.isfile(cmdargs[0]):
            if os.path.isfile(os.path.join('system', cmdargs[0])):
                cmdargs[0] = os.path.join('system', cmdargs[0])
            else:
                lg.warn('can not find executable file ' + cmdargs[0])
                return None

    cmdargs += args_list

    if _Debug:
        lg.out(_DebugLevel, 'run_upnpc.run is going to execute: %s' % cmdargs)

    try:
        out_data, returncode = execute_in_shell(cmdargs)
    except:
        lg.exc()
        return None

    if _Debug:
        lg.out(_DebugLevel, '    %s finished with return code: %s' % (str(_CurrentProcess), str(returncode)))
    _CurrentProcess = None

    return out_data


#------------------------------------------------------------------------------


def info():
    cmd_out = run(('-l', ))
    if cmd_out is None:
        return None, None, None
    regexp1 = '^\s*(\d+)\s*(\w+)\s*(\d+)->(\d+\.\d+\.\d+\.\d+):(\d+)\s+(.+)$'
    regexp2 = '^Local LAN ip address : (\d+\.\d+\.\d+\.\d+).*$'
    regexp3 = '^ExternalIPAddress = (\d+\.\d+\.\d+\.\d+).*$'
    regexp4 = 'No IGD UPnP Device found on the network'
    upnp_device_not_found = re.search(regexp4, cmd_out, re.MULTILINE)
    if upnp_device_not_found is not None:
        return None, None, None
    l = []
    for i in re.findall(regexp1, cmd_out, re.MULTILINE):
        try:
            l.append((int(i[2]), i[3], i[1], i[5]))
        except:
            continue
    search_local_ip = re.search(regexp2, cmd_out, re.MULTILINE)
    search_external_ip = re.search(regexp3, cmd_out, re.MULTILINE)
    local_ip = ''
    external_ip = ''
    if search_local_ip is not None:
        local_ip = search_local_ip.group(1)
    if search_external_ip is not None:
        external_ip = search_external_ip.group(1)
    return local_ip, external_ip, l


#------------------------------------------------------------------------------


def lst():
    cmd_out = run(('-l', ))
    if cmd_out is None:
        return None
    regexp1 = '^\s*(\d+)\s*(\w+)\s*(\d+)->(\d+\.\d+\.\d+\.\d+):(\d+)\s+(.+)$'
    regexp4 = 'No IGD UPnP Device found on the network'
    upnp_device_not_found = re.search(regexp4, cmd_out, re.MULTILINE)
    if upnp_device_not_found is not None:
        return None
    l = []
    for i in re.findall(regexp1, cmd_out, re.MULTILINE):
        try:
            #         port       ip    proto  decription
            l.append((int(i[2]), i[3], i[1], i[5]))
        except:
            continue
    return l


#------------------------------------------------------------------------------


def add(port, proto):
    global _MyPortMapping
    # cmd_out = run(('-r', str(port), str(proto)))
    cmd_out = run(('-e', '"BitDust TCP on port %s"' % (port), '-r', str(port), str(proto)))
    if cmd_out is None:
        return None
    _MyPortMapping[str(port)] = str(proto)
    return cmd_out


#------------------------------------------------------------------------------


def dlt(port, proto):
    global _MyPortMapping
    cmd_out = run(('-d', str(port), str(proto)))
    if cmd_out is None:
        return None
    _MyPortMapping.pop(str(port), '')
    return cmd_out


#------------------------------------------------------------------------------


def clear():
    s = ''
    l = lst()
    for i in l:
        s += str(dlt(i[0], i[2])) + '\n'
    return s


#------------------------------------------------------------------------------


def update(requested_port, attempt=0, new_port=-1):
    global _MyPortMapping
    global _LastUpdateResultDict
    port = requested_port
    requested_port_busy = False
    if _Debug:
        lg.out(_DebugLevel, 'run_upnpc.update %s attempt=%s new_port=%s' % (str(port), str(attempt), str(new_port)))

    local_ip, external_ip, port_map = info()

    if local_ip is None or external_ip is None or port_map is None:
        _LastUpdateResultDict[port] = 'upnp-not-found'
        return 'upnp-not-found', port

    local_ports = {}
    for i in port_map:
        if str(i[0]).strip() == str(requested_port):
            requested_port_busy = True
        if i[1] == local_ip and (str(i[3]).find('libminiupnpc') >= 0 or str(i[3]).find('BitDust') >= 0):
            local_ports[i[0]] = (i[2], i[3])

    if _Debug:
        lg.out(_DebugLevel, 'run_upnpc.update requested_port_busy=%s local_ports=%s' % (requested_port_busy, str(list(local_ports.keys()))))

    if int(port) in list(local_ports.keys()):
        _MyPortMapping[str(port)] = 'TCP'
        if _Debug:
            lg.out(_DebugLevel, 'run_upnpc.update PORT %s mapped SUCCESSFULLY!!! all port maps: %s' % (str(port), str(list(_MyPortMapping.keys()))))
        _LastUpdateResultDict[port] = 'upnp-done'
        return 'upnp-done', port

    if int(new_port) > 0 and int(new_port) in list(local_ports.keys()):
        _MyPortMapping[str(new_port)] = 'TCP'
        if _Debug:
            lg.out(_DebugLevel, 'run_upnpc.update NEW PORT %s mapped SUCCESSFULLY!!! all port maps: %s' % (str(port), str(list(_MyPortMapping.keys()))))
        _LastUpdateResultDict[new_port] = 'upnp-done'
        return 'upnp-done', new_port

    if attempt == 0:
        add(port, 'TCP')

    elif attempt == 1:
        closest_port = -1
        closest_value = 99999999
        for p in local_ports.keys():
            v = abs(int(p) - int(port))
            if v < closest_value:
                closest_value = v
                closest_port = p
        if closest_port >= 0:
            dlt(closest_port, 'TCP')
            if requested_port_busy:
                new_port = int(port) + random.randint(1, 100)
                add(new_port, 'TCP')
                port = new_port
                new_port = -1
            else:
                add(port, 'TCP')
        else:
            new_port = int(port) + random.randint(1, 100)
            add(new_port, 'TCP')
            port = new_port
            new_port = -1

    else:
        _LastUpdateResultDict[port] = 'upnp-error'
        return 'upnp-error', port

    result, port = update(port, attempt + 1, new_port)
    _LastUpdateResultDict[port] = result
    return result, port


#------------------------------------------------------------------------------


def last_result(proto):
    global _LastUpdateResultDict
    return _LastUpdateResultDict.get(proto, 'upnp-no-info')


#-------------------------------------------------------------------------------


def main():
    import pprint
    lg.set_debug_level(14)
    if sys.argv.count('list'):
        maps = lst()
        for itm in maps:
            pprint.pprint(itm)
    elif sys.argv.count('info'):
        locip, extip, maps = info()
        pprint.pprint(locip)
        pprint.pprint(extip)
        for itm in maps:
            pprint.pprint(itm)
    elif sys.argv.count('add'):
        print(add(sys.argv[2], 'TCP'))
    elif sys.argv.count('del'):
        print(dlt(sys.argv[2], 'TCP'))
    elif sys.argv.count('update'):
        bpio.init()
        settings.init()
        init()
        pprint.pprint(update(sys.argv[2]))
        settings.shutdown()
    elif sys.argv.count('clear'):
        print(clear())
    else:
        print('usage:')
        print('run_upnpc.py info')
        print('run_upnpc.py list')
        print('run_upnpc.py add [port]')
        print('run_upnpc.py del [port]')
        print('run_upnpc.py update [port]')
        print('run_upnpc.py clear')


if __name__ == '__main__':
    main()



#todo: make registrations produce mirror hashes
#todo: make sure registrations newer than latest block are ignored
#todo: rollbacks inside node; make sure delagete/ip is only allowed characters

#operation: staking:register
#last working integrated implementation: https://github.com/bismuthfoundation/Bismuth/commit/69fd3aafd31cce6def1b09ffc64c95d457243c8a

import sqlite3
import log
from quantizer import quantize_eight
import mempool as mp
from hashlib import blake2b
from essentials import percentage


def execute_param(cursor, query, param):
    """Secure execute w/ param for slow nodes"""
    while True:
        try:
            cursor.execute(query, param)
            break
        except Exception as e:
            app_log.warning('Database query: {} {} {}'.format(cursor, query, param))
            app_log.warning('Database retry reason: {}'.format(e))
    return cursor


def balanceget_at_block(balance_address, block, h3):
    # verify balance

    credit_ledger = Decimal('0')
    for entry in execute_param(
        h3,
        ('SELECT amount FROM transactions WHERE block_height <= ? AND block_height >= ? AND recipient = ?;'),
        (
            block,
            -block,
            balance_address,
        ),
    ):
        try:
            credit_ledger = quantize_eight(credit_ledger) + quantize_eight(entry[0])
            credit_ledger = 0 if credit_ledger is None else credit_ledger
        except:
            credit_ledger = 0

    fees = Decimal('0')
    debit_ledger = Decimal('0')

    for entry in execute_param(
        h3,
        ('SELECT fee, amount FROM transactions WHERE block_height <= ? AND block_height >= ? AND address = ?;'),
        (
            block,
            -block,
            balance_address,
        ),
    ):
        try:
            fees = quantize_eight(fees) + quantize_eight(entry[0])
            fees = 0 if fees is None else fees
        except:
            fees = 0

        try:
            debit_ledger = debit_ledger + Decimal(entry[1])
            debit_ledger = 0 if debit_ledger is None else debit_ledger
        except:
            debit_ledger = 0

    debit = quantize_eight(debit_ledger)

    rewards = Decimal('0')
    for entry in execute_param(
        h3,
        ('SELECT reward FROM transactions WHERE block_height <= ? AND block_height >= ? AND recipient = ?;'),
        (
            block,
            -block,
            balance_address,
        ),
    ):
        try:
            rewards = quantize_eight(rewards) + quantize_eight(entry[0])
            rewards = 0 if rewards is None else rewards
        except:
            rewards = 0

    balance = quantize_eight(credit_ledger - debit - fees + rewards)
    # app_log.info("Mempool: Projected transction address balance: " + str(balance))
    return str(balance)  #, str (credit_ledger), str (debit), str (fees), str (rewards)


def check_db(index, index_cursor):
    index_cursor.execute('CREATE TABLE IF NOT EXISTS staking (block_height INTEGER, timestamp NUMERIC, address, balance)')
    index.commit()


def staking_update(conn, c, index, index_cursor, mode, reg_phase_end, app_log):
    """update register of staking based on the current phase (10000 block intervals)"""
    if mode not in ('normal', 'reindex'):
        raise ValueError('Wrong value for staking_update function')

    check_db(index, index_cursor)

    if mode == 'reindex':
        app_log.warning('staking database will be reindexed')
        index_cursor.execute('DELETE FROM staking')
        index.commit()

    reg_phase_start = reg_phase_end - 10000
    app_log.warning('reg_phase_start: {}'.format(reg_phase_start))
    app_log.warning('reg_phase_end: {}'.format(reg_phase_end))

    c.execute(
        'SELECT block_height, timestamp, address, recipient,operation, openfield FROM transactions WHERE block_height >= ? AND block_height <= ? AND operation = ? ORDER BY block_height, timestamp LIMIT 100', (
            reg_phase_start,
            reg_phase_end,
            'staking:register',
        )
    )
    results = c.fetchall()  #more efficient than "for row in"

    for row in results:
        try:
            block_height = row[0]
            timestamp = row[1]
            address = row[2]

            try:
                index_cursor.execute('SELECT * from staking WHERE address = ?', (address, ))
                dummy = index_cursor.fetchall()[0]  #check for uniqueness
                app_log.warning('staking already registered: {}'.format(address))
            except:
                app_log.warning('address: {}'.format(address))
                balance = balanceget_at_block(address, reg_phase_end, c)

                if quantize_eight(balance) >= 10000:
                    index_cursor.execute('INSERT INTO staking VALUES (?, ?, ?, ?)', (block_height, timestamp, address, balance))
                    index.commit()

                    app_log.warning('Staking added: {} {}'.format(block_height, address))
                else:
                    app_log.warning('Insufficient balance for staking')
        except Exception as e:
            app_log.warning('Staking registration ran into the following issue: {}'.format(e))

    return reg_phase_start, reg_phase_end


def mirror_hash_generate(c):
    # new hash
    c.execute('SELECT * FROM transactions WHERE block_height = (SELECT block_height FROM transactions ORDER BY block_height ASC LIMIT 1)')
    result = c.fetchall()
    mirror_hash = blake2b(str(result).encode(), digest_size=20).hexdigest()
    return mirror_hash
    # new hash


def staking_payout(conn, c, index, index_cursor, block_height, timestamp, app_log):
    'payout, to be run every 10k blocks'

    index_cursor.execute('SELECT * FROM staking')
    staking = index_cursor.fetchall()
    app_log.warning('staking: {}'.format(staking))
    staking_total = len(staking)

    mirror_hash = mirror_hash_generate(c)

    for staking in staking:
        block_staking = staking[0]
        if block_staking <= block_height:
            address = staking[2]
            balance_savings = staking[3]
            app_log.warning('balance_savings: {}'.format(balance_savings))
            stake = str(quantize_eight(percentage(100/staking_total, balance_savings)))
            app_log.warning('stake: {}'.format(stake))

            try:
                c.execute('SELECT * from transactions WHERE block_height = ? AND recipient = ?', (
                    -block_height,
                    address,
                ))
                dummy = c.fetchall()[0]  # check for uniqueness
                app_log.warning('staking payout already processed: {} {}'.format(block_height, address))

            except:
                """skip direct bis payouts
                c.execute("INSERT INTO transactions VALUES (?,?,?,?,?,?,?,?,?,?,?,?)",(-block_height,timestamp,"staking",address,stake,"0","0",mirror_hash,"0","0","mnpayout","0"))
                conn.commit()
                app_log.warning ("staking payout added: {} {}".format (block_height, address))
                """

                #fuel
                stake_int = int(float(stake))
                if stake_int < 1:
                    stake_int = 1

                c.execute('INSERT INTO transactions VALUES (?,?,?,?,?,?,?,?,?,?,?,?)', (-block_height, timestamp, 'staking', address, '0', '0', '0', mirror_hash, '0', '0', 'token:transfer', 'fuel:{}'.format(stake_int)))
                conn.commit()
                app_log.warning('Staking fuel payout added: {} {}'.format(block_height, address))
                #fuel
        else:
            app_log.warning('staking is registered ahead of current block')


def staking_revalidate(conn, c, index, index_cursor, block, app_log):
    'remove nodes that removed balance, to be run every 10k blocks'

    index_cursor.execute('SELECT * FROM staking')
    staking = index_cursor.fetchall()

    for staking in staking:
        app_log.warning(staking)
        address = staking[2]
        app_log.warning('address: {}'.format(address))
        balance_savings = staking[3]
        app_log.warning('balance_savings: {}'.format(balance_savings))
        balance = balanceget_at_block(address, block, c)
        app_log.warning('balance: {}'.format(balance))

        if quantize_eight(balance) < 10000:
            index_cursor.execute('DELETE FROM staking WHERE address = ?', (address, ))
            index.commit()
        else:  #update balance
            index_cursor.execute('UPDATE staking SET balance = ? WHERE address = ?', (balance, address))
            index.commit()
            app_log.warning('staking balance updated from {} to {} for {}'.format(balance_savings, balance, address))


if __name__ == '__main__':

    import options
    config = options.Get()
    config.read()

    app_log = log.log('solvency.log', 'WARNING', True)
    mp.MEMPOOL = mp.Mempool(app_log, config, None, False)

    conn = sqlite3.connect('static/ledger_test.db')
    conn.text_factory = str
    c = conn.cursor()

    index = sqlite3.connect('static/index_test.db')
    index.text_factory = str
    index_cursor = index.cursor()

    address = '4edadac9093d9326ee4b17f869b14f1a2534f96f9c5d7b48dc9acaed'
    staking_update(conn, c, index, index_cursor, 'normal', 736600, app_log)
    staking_payout(conn, c, index, index_cursor, 736600, 1525304875, app_log)
    staking_revalidate(conn, c, index, index_cursor, 736600, app_log)



def default_network_info(): return {
  "service_bismuth_blockchain": {
    "initial_difficulty": 10,
    "genesis_address": "cc846806ba14b8ec5a042d254beeeb637ee91033fa0f84c66063e0a9",
    "foundation_miners": [
        "884dbabd8e214cf1ac80c0718d6c66fd9cdb7a28cfdd0d2bc3b053fe",
        "d6d7e856b50ea783c388e17137eb789dd21a08295bbaa834a4b9b9ea"
    ],
    "known_nodes": [
      {
        "host": "p2p-id.ru",
        "node_tcp_port": 15658,
        "mining_pool_tcp_port": 18525,
        "explorer_http_port": 19080
      },
      {
        "host": "bahamas.ai",
        "node_tcp_port": 15658,
        "mining_pool_tcp_port": 18525,
        "explorer_http_port": 19080
      }
    ]
  },
  "service_entangled_dht": {
    "bucket_size": 4,
    "default_age": 43200,
    "known_nodes": [
      {
        "host": "p2p-id.ru",
        "udp_port": 14441
      },
      {
        "host": "p2p-alice.ai",
        "udp_port": 14441
      },
      {
        "host": "bitdust.ai",
        "udp_port": 14441
      },
      {
        "host": "root-node.ai",
        "udp_port": 14441
      },
      {
        "host": "bahamas.ai",
        "udp_port": 14441
      },
      {
        "host": "peer2peer.ai",
        "udp_port": 14441
      },
      {
        "host": "seed.bitdust.io",
        "udp_port": 14441
      }
    ],
    "max_age": 86400,
    "parallel_calls": 4,
    "refresh_timeout": 600,
    "rpc_timeout": 10
  },
  "service_gateway": {
    "session_key_type": "AES"
  },
  "service_identity_propagate": {
    "known_servers": [
      {
        "host": "p2p-id.ru",
        "http_port": 80
      },
      {
        "host": "p2p-alice.ai",
        "http_port": 80
      },
      {
        "host": "bitdust.ai",
        "http_port": 80
      },
      {
        "host": "root-node.ai",
        "http_port": 80
      },
      {
        "host": "bahamas.ai",
        "http_port": 80
      },
      {
        "host": "peer2peer.ai",
        "http_port": 80
      },
      {
        "host": "seed.bitdust.io",
        "http_port": 80
      }
    ],
    "whitelisted_servers": []
  }
}



#!/usr/bin/python
# my_id.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (my_id.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: my_id
"""

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 8

#------------------------------------------------------------------------------

import os
import sys
import time
import tempfile

#------------------------------------------------------------------------------

if __name__ == '__main__':
    import os.path as _p
    sys.path.insert(0, _p.abspath(_p.join(_p.dirname(_p.abspath(sys.argv[0])), '..')))

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.system import bpio
from bitdust.system import local_fs

from bitdust.main import settings
from bitdust.main import events

from bitdust.lib import misc
from bitdust.lib import nameurl
from bitdust.lib import strng

from bitdust.crypt import key

from bitdust.userid import identity
from bitdust.userid import id_url

#------------------------------------------------------------------------------

_LocalIdentity = None
_LocalIDURL = None
_LocalID = None
_LocalName = None
_ValidTransports = [
    'tcp',
    'udp',
    'http',
    'proxy',
]

#------------------------------------------------------------------------------


def init():
    """
    Will be called in main thread at start up.

    Can put here some minor things if needed.
    """
    if _Debug:
        lg.out(_DebugLevel, 'my_id.init')
    loadLocalIdentity()


def shutdown():
    if _Debug:
        lg.out(_DebugLevel, 'my_id.shutdown')
    forgetLocalIdentity()


#-------------------------------------------------------------------------------


def isLocalIdentityExists():
    """
    Return True if local file `~/.bitdust/[network name]/metadata/localidentity` exists.
    """
    return os.path.isfile(settings.LocalIdentityFilename())


def isLocalIdentityReady():
    """
    Return True if local identity object already initialized and stored in
    memory.
    """
    global _LocalIdentity
    return _LocalIdentity is not None


def setLocalIdentity(ident):
    """
    Set local identity object in the memory.
    """
    global _LocalIdentity
    global _LocalIDURL
    global _LocalID
    global _LocalName
    if not ident:
        return
    modified = False
    old_json = {}
    new_json = {}
    if _LocalIdentity:
        current_src = _LocalIdentity.serialize()
        if current_src != ident.serialize():
            modified = True
            li_json = _LocalIdentity.serialize_json()
            old_json['revision'] = li_json['revision']
            old_json['contacts'] = li_json['contacts']
    _LocalIdentity = ident
    _LocalIDURL = _LocalIdentity.getIDURL()
    _LocalIDURL.refresh(replace_original=True)
    _LocalID = _LocalIDURL.to_id()
    _LocalName = _LocalIdentity.getIDName()
    li_json = _LocalIdentity.serialize_json()
    new_json['revision'] = li_json['revision']
    new_json['contacts'] = li_json['contacts']
    if modified:
        events.send('local-identity-modified', data=dict(old=old_json, new=new_json))
    else:
        events.send('local-identity-set', data=new_json)
    try:
        id_url.identity_cached(_LocalIdentity)
    except:
        lg.exc()


def setLocalIdentityXML(idxml):
    """
    Construct identity object from XML string and save it to the memory.
    """
    setLocalIdentity(identity.identity(xmlsrc=idxml))


def getLocalIdentity():
    """
    Returns my identity object.
    """
    global _LocalIdentity
    if not isLocalIdentityReady():
        loadLocalIdentity()
    return _LocalIdentity


def getIDName():
    """
    Returns my account name, this is a filename part of IDURL without '.xml'.
    """
    global _LocalName
    if _LocalName is None:
        localIdent = getLocalIdentity()
        if localIdent:
            _LocalName = localIdent.getIDName()
    return _LocalName


def getGlobalID(key_alias=None):
    """
    Returns my global user id - according to my current IDURL.
    """
    global _LocalID
    if not key_alias and _LocalID is not None:
        return _LocalID
    if key_alias == 'master' and _LocalID is not None:
        return strng.to_text('{}${}'.format(key_alias, _LocalID))
    from bitdust.userid import global_id
    glob_id = global_id.UrlToGlobalID(getIDURL())
    if not glob_id:
        return glob_id
    if key_alias:
        glob_id = strng.to_text('{}${}'.format(key_alias, glob_id))
    return glob_id


def getIDURL():
    """
    Returns my IDURL as a field.
    """
    global _LocalIDURL
    if _LocalIDURL is None:
        localIdent = getLocalIdentity()
        if localIdent:
            _LocalIDURL = localIdent.getIDURL()
    return _LocalIDURL


def getID():
    """
    Returns my global ID as a string.
    """
    return getGlobalID()


#------------------------------------------------------------------------------


def loadLocalIdentity():
    """
    The file `~/.bitdust/[network name]/metadata/localidentity` keeps the user
    identity in XML format.
    The method reads the local file and set into object in memory.
    """
    global _LocalIdentity
    xmlid = ''
    lid_filename = bpio.portablePath(settings.LocalIdentityFilename())
    key_filename = settings.KeyFileName()
    if not os.path.isfile(key_filename):
        keyfilenamelocation = settings.KeyFileNameLocation()
        if not os.path.isfile(keyfilenamelocation):
            lg.warn('not possible to load local identity because master key file does not exist at specified location')
            return False
        key_filename = bpio.ReadTextFile(keyfilenamelocation).strip()
        if not os.path.isfile(key_filename):
            lg.warn('not possible to load local identity because master key file does not exist at default location')
            return False
    if os.path.exists(lid_filename):
        xmlid = bpio.ReadTextFile(lid_filename)
        if _Debug:
            lg.out(_DebugLevel, 'my_id.loadLocalIdentity %d bytes read from %s' % (len(xmlid), lid_filename))
    if not xmlid:
        if _Debug:
            lg.out(_DebugLevel, 'my_id.loadLocalIdentity SKIPPED, local identity in %s is EMPTY !!!' % lid_filename)
        return False
    lid = identity.identity(xmlsrc=xmlid)
    if not lid.isCorrect():
        if _Debug:
            lg.out(_DebugLevel, 'my_id.loadLocalIdentity ERROR loaded identity is not Correct')
        return False
    if not lid.Valid():
        if _Debug:
            lg.out(_DebugLevel, 'my_id.loadLocalIdentity ERROR loaded identity is not Valid')
        return False
    setLocalIdentity(lid)
    setTransportOrder(getOrderFromContacts(_LocalIdentity))
    if _Debug:
        lg.out(_DebugLevel, 'my_id.loadLocalIdentity my global id is %s' % getGlobalID())
    return True


def saveLocalIdentity():
    """
    Save identity object from memory into local file.

    Do sign the identity than serialize to write to the file.
    """
    global _LocalIdentity
    if not isLocalIdentityReady():
        lg.warn('ERROR local identity not exist!')
        return False
    if not _LocalIdentity.isCorrect():
        lg.warn('local identity is not correct')
        return False
    _LocalIdentity.sign()
    if not _LocalIdentity.Valid():
        lg.err('local identity is not valid')
        return False
    xmlid = _LocalIdentity.serialize(as_text=True)
    filename = bpio.portablePath(settings.LocalIdentityFilename())
    bpio.WriteTextFile(filename, xmlid)
    setTransportOrder(getOrderFromContacts(_LocalIdentity))
    events.send('local-identity-written', data=dict(idurl=_LocalIdentity.getIDURL(), filename=filename))
    if _Debug:
        lg.out(_DebugLevel, 'my_id.saveLocalIdentity %d bytes wrote to %s' % (len(xmlid), filename))
    return True


def forgetLocalIdentity():
    global _LocalIdentity
    global _LocalIDURL
    global _LocalID
    global _LocalName
    if not isLocalIdentityReady():
        if _Debug:
            lg.out(_DebugLevel, 'my_id.forgetLocalIdentity ERROR local identity not exist!')
        return False
    if _Debug:
        lg.out(_DebugLevel, 'my_id.saveLocalIdentity')
    _LocalIdentity = None
    _LocalIDURL = None
    _LocalID = None
    _LocalName = None
    events.send('local-identity-cleaned', data=dict())
    return True


def eraseLocalIdentity(do_backup=True):
    if do_backup:
        if os.path.isfile(settings.LocalIdentityFilename()):
            current_identity_xmlsrc = local_fs.ReadBinaryFile(settings.LocalIdentityFilename())
            if current_identity_xmlsrc:
                fd, fname = tempfile.mkstemp(prefix='localidentity_', dir=settings.MetaDataDir())
                os.write(fd, current_identity_xmlsrc)
                os.close(fd)
                lg.info('created backup copy of my local identity in the file : %r' % fname)
    filename = bpio.portablePath(settings.LocalIdentityFilename())
    if not os.path.exists(filename):
        if _Debug:
            lg.out(_DebugLevel, 'my_id.eraseLocalIdentity SKIP file %s not exist' % filename)
        return True
    if not os.path.isfile(filename):
        if _Debug:
            lg.out(_DebugLevel, 'my_id.eraseLocalIdentity ERROR path %s is not a file' % filename)
        return False
    try:
        os.remove(filename)
    except:
        lg.exc()
        return False
    events.send('local-identity-erased', data=dict())
    if _Debug:
        lg.out(_DebugLevel, 'my_id.eraseLocalIdentity file %s was deleted' % filename)
    return True


#------------------------------------------------------------------------------


def getValidTransports():
    global _ValidTransports
    return _ValidTransports


def isValidTransport(transport):
    """
    Check string to be a valid transport.

    See ``lib.transport_control' for more details.
    """
    global _ValidTransports
    if transport in _ValidTransports:
        return True
    else:
        return False


def validateTransports(orderL):
    """
    Validate a list of strings - all must be a valid transports.
    """
    global _ValidTransports
    transports = []
    for transport in orderL:
        if isValidTransport(transport):
            transports.append(transport)
        else:
            lg.warn('invalid entry in transport list: %s , ignored' % str(transport))
    if len(transports) == 0:
        if _Debug:
            lg.out(_DebugLevel, 'my_id.validateTransports ERROR no valid transports, using default transports ' + str(_ValidTransports))
        transports = _ValidTransports


#    if len(transports) != len(orderL):
#        lg.out(1, 'my_id.validateTransports ERROR Transports contained an invalid entry, need to figure out where it came from.')
    return transports


def setTransportOrder(orderL):
    """
    Validate transports and save the list in the `~/.bitdust/[network name]/metadata/torder`.
    It is useful to remember the priority of used transports.
    """
    orderl = orderL
    orderL = validateTransports(orderL)
    orderTxt = ' '.join(orderl)
    if _Debug:
        lg.out(_DebugLevel, 'my_id.setTransportOrder: ' + str(orderTxt))
    bpio.WriteTextFile(settings.DefaultTransportOrderFilename(), orderTxt)


def getTransportOrder():
    """
    Read and validate tranports from `~/.bitdust/[network name]/metadata/torder` file.
    """
    global _ValidTransports
    if _Debug:
        lg.out(_DebugLevel, 'my_id.getTransportOrder')
    order = bpio.ReadTextFile(settings.DefaultTransportOrderFilename()).strip()
    if order == '':
        orderL = _ValidTransports
    else:
        orderL = order.split(' ')
        orderL = validateTransports(orderL)
    setTransportOrder(orderL)
    return orderL


def getOrderFromContacts(ident):
    """
    A wrapper for ``identity.getProtoOrder`` method.
    """
    return ident.getProtoOrder()


#------------------------------------------------------------------------------


def buildProtoContacts(id_obj, skip_transports=[]):
    """
    Create a full list of needed transport methods to be able to accept
    incoming traffic from other nodes.

    Make calls to transport services to build a list of my contacts.
    """
    from bitdust.services import driver
    # prepare contacts
    current_contats = id_obj.getContactsByProto()
    current_order = id_obj.getProtoOrder()
    if _Debug:
        lg.out(_DebugLevel, 'my_id.buildProtoContacts')
        lg.out(_DebugLevel, '    current contacts: %s' % str(current_contats))
        lg.out(_DebugLevel, '    current order: %s' % str(current_order))
    new_contacts = {}
    new_order_correct = []
    # prepare list of active transports
    active_transports = []
    for proto in getValidTransports():
        if proto in skip_transports:
            continue
        if not settings.transportIsEnabled(proto):
            continue
        if not settings.transportReceivingIsEnabled(proto):
            continue
        if not driver.is_on('service_%s_transport' % proto):
            lg.warn('transport "%s" is enabled, but service_%s_transport() is not ready yet' % (proto, proto))
            continue
        active_transports.append(proto)
    # sort active transports by priority
    if _Debug:
        lg.out(_DebugLevel, '    active transports: %s' % str(active_transports))
    active_transports.sort(key=settings.getTransportPriority)
    if _Debug:
        lg.out(_DebugLevel, '    sorted transports: %s' % str(active_transports))
    if not driver.is_on('service_gateway'):
        new_contacts = current_contats
        new_order_correct = current_order
        lg.warn('service_gateway() is not started, use my current contacts as a source')
    else:
        from bitdust.transport import gateway
        # build contacts data according transports priorities
        new_order = current_order
        for proto in active_transports:
            clist = gateway.transport(proto).interface.build_contacts(id_obj)
            cdict = {}
            corder = []
            for contact in clist:
                cproto, _ = contact.split(b'://')
                cdict[cproto] = contact
                corder.append(cproto)
            new_contacts.update(cdict)
            for cproto in corder:
                if cproto not in new_order:
                    new_order.append(cproto)
        new_order_correct = list(new_order)
        for nproto in new_order:
            if nproto not in list(new_contacts.keys()):
                new_order_correct.remove(nproto)
    if _Debug:
        lg.out(_DebugLevel, '    new contacts: %s' % str(new_contacts))
        lg.out(_DebugLevel, '    new order: %s' % str(new_order_correct))
    return new_contacts, new_order_correct


def buildDefaultIdentity(name='', ip='', idurls=[], revision=0):
    """
    Use some local settings and config files to create some new identity.

    Nice to provide a user name or it will have a form like: [ip_address]_[date].
    """
    if not ip:
        ip = misc.readExternalIP()
    if not name:
        name = ip.replace('.', '-') + '_' + time.strftime('%M%S')
    if _Debug:
        lg.args(_DebugLevel, name=name, ip=ip, idurls=idurls)
    # this is my IDURL address
    # you can have many IDURL locations for same identity
    # just need to keep all them synchronized
    # this is identity propagate procedure, see p2p/propagate.py
    if len(idurls) == 0:
        idurls.append(b'http://127.0.0.1/%s.xml' % strng.to_bin(name.lower()))
    # create a new identity object
    # it is stored in memory and another copy will be located on disk drive
    ident = identity.identity(xmlsrc=identity.default_identity_src)
    ident.setSources(idurls)
    # create a full list of needed transport methods
    # to be able to accept incoming traffic from other nodes
    new_contacts, new_order = buildProtoContacts(ident)
    if len(new_contacts) == 0:
        if settings.enableTCP() and settings.enableTCPreceiving():
            new_contacts['tcp'] = b'tcp://' + strng.to_bin(ip) + b':' + strng.to_bin(str(settings.getTCPPort()))
            new_order.append('tcp')
        if settings.enableUDP() and settings.enableUDPreceiving():
            _, servername, _, _ = nameurl.UrlParse(ident.getSources(as_originals=True)[0])
            new_contacts['udp'] = b'udp://' + strng.to_bin(name.lower()) + b'@' + strng.to_bin(servername)
            new_order.append('udp')
        if settings.enableHTTP() and settings.enableHTTPreceiving():
            new_contacts['http'] = b'http://' + strng.to_bin(ip) + b':' + strng.to_bin(str(settings.getHTTPPort()))
            new_order.append('http')
    # erase current contacts from my identity
    ident.clearContacts()
    # add contacts data to the local identity
    for proto in new_order:
        contact = new_contacts.get(proto, None)
        if contact is None:
            lg.warn('proto %s was not found in contacts' % proto)
            continue
        ident.setProtoContact(proto, contact)
    # set other info
    # ident.certificates = []
    ident.setDate(time.strftime('%b %d, %Y'))
    ident.setPostage(1)
    ident.setRevision(revision)
    ident.setVersion('')  # TODO: put latest git commit hash here
    # update software version number
    # version_number = bpio.ReadTextFile(settings.VersionNumberFile()).strip()
    # repo, location = misc.ReadRepoLocation()
    # ident.version = (version_number.strip() + ' ' + repo.strip() + ' ' + bpio.osinfo().strip()).strip()
    # build a version info
    # vernum = bpio.ReadTextFile(settings.VersionNumberFile())
    # repo, location = misc.ReadRepoLocation()
    # ident.version = (vernum.strip() + ' ' + repo.strip() + ' ' + bpio.osinfo().strip()).strip()
    # put my public key in my identity
    ident.setPublicKey(key.MyPublicKey())
    # generate signature
    ident.sign()
    # validate new identity
    if not ident.Valid():
        lg.warn('generated identity is not valid !!!')
    return ident


def rebuildLocalIdentity(identity_object=None, skip_transports=[], new_sources=None, revision_up=False, new_revision=None, save_identity=True):
    """
    If some transports was enabled or disabled we want to update identity
    contacts. Just empty all of the contacts and create it again in the same
    order.

    Also increase revision number by one - others may keep track of my modifications.
    """
    # remember the current identity - full XML source code
    current_identity_xmlsrc = getLocalIdentity().serialize()
    if _Debug:
        lg.out(_DebugLevel, 'my_id.rebuildLocalIdentity current identity is %d bytes long new_revision=%r' % (len(current_identity_xmlsrc), new_revision))
    # getting a copy of local identity to be modified or another object to be used
    lid = identity_object or identity.identity(xmlsrc=current_identity_xmlsrc)
    # create a full list of needed transport methods
    # to be able to accept incoming traffic from other nodes
    new_contacts, new_order = buildProtoContacts(lid, skip_transports=skip_transports)
    # erase current contacts from my identity
    lid.clearContacts()
    # add contacts data to the local identity
    lid.setContactsFromDict(new_contacts, new_order)
    # if I need to rotate my sources do it now
    if new_sources:
        lid.setSources(new_sources)
    # update software version number
    # TODO: need to read GIT commit hash here instead of version
    vernum = strng.to_bin(bpio.ReadTextFile(settings.VersionNumberFile())).strip()
    # repo, _ = misc.ReadRepoLocation()
    repo = 'sources'
    # lid.setVersion((vernum + b' ' + strng.to_bin(repo.strip()) + b' ' + strng.to_bin(bpio.osinfo().strip()).strip()))
    # TODO: add latest commit hash from the GIT repo to the version
    lid.setVersion(vernum + b' ' + strng.to_bin(repo.strip()))
    # generate signature with changed content
    lid.sign()
    new_xmlsrc = lid.serialize()
    changed = False
    if new_xmlsrc != current_identity_xmlsrc or revision_up or new_revision:
        if not new_revision:
            new_revision = lid.getRevisionValue() + 1
        try:
            lid.setRevision(new_revision)
        except:
            lg.exc()
            return False
        # generate signature again because revision were changed !!!
        lid.sign()
        lg.info('incremented my identity revision: %d' % lid.getRevisionValue())
        changed = True
    else:
        # no modifications in my identity - cool !!!
        if _Debug:
            lg.out(_DebugLevel, '    same revision: %d' % lid.getRevisionValue())
    if _Debug:
        lg.out(_DebugLevel, '    version: %r' % lid.version)
        lg.out(_DebugLevel, '    contacts: %r' % lid.contacts)
        lg.out(_DebugLevel, '    sources: %r' % lid.getSources(as_originals=True))
    if changed:
        if save_identity:
            # finally saving modified local identity
            if _Debug:
                lg.out(_DebugLevel, '    SAVING new identity #%s' % lid.revision)
            # remember the new identity
            setLocalIdentity(lid)
            saveLocalIdentity()
            # NOW TEST IT!
            # forgetLocalIdentity()
            # loadLocalIdentity()
            if _Debug:
                lg.out(_DebugLevel, '    LOCAL IDENTITY CORRECT: %r' % getLocalIdentity().isCorrect())
                lg.out(_DebugLevel, '    LOCAL IDENTITY VALID: %r' % getLocalIdentity().Valid())
    lg.info('my identity HAS %sBEEN changed' % (('' if changed else 'NOT ')))
    if _Debug:
        lg.out(_DebugLevel, '\n' + strng.to_text(lid.serialize()) + '\n')
    return changed
